"""
This module includes AfeSettingsValidator class.
"""
from typing import List, Dict

import pandas as pd

from organon.afe.core.businessobjects.afe_static_objects import AfeStaticObjects
from organon.afe.domain.enums.afe_date_column_type import AfeDateColumnType
from organon.afe.domain.enums.afe_learning_type import AfeLearningType
from organon.afe.domain.enums.afe_target_column_type import AfeTargetColumnType
from organon.afe.domain.enums.date_resolution import DateResolution
from organon.afe.domain.enums.record_source_type import RecordSourceType
from organon.afe.domain.modelling.supervised.afe_supervised_algorithm_settings import AfeSupervisedAlgorithmSettings
from organon.afe.domain.modelling.unsupervised.afe_unsupervised_algorithm_settings import \
    AfeUnsupervisedAlgorithmSettings
from organon.afe.domain.reporting.base_afe_model_output import BaseAfeModelOutput
from organon.afe.domain.settings.afe_algorithm_settings import AfeAlgorithmSettings
from organon.afe.domain.settings.afe_date_column import AfeDateColumn
from organon.afe.domain.settings.afe_output_settings import AfeOutputSettings
from organon.afe.domain.settings.afe_settings_type_validators import validate_types_modelling_settings, \
    validate_types_scoring_settings
from organon.afe.domain.settings.afe_target_column import AfeTargetColumn
from organon.afe.domain.settings.base_afe_data_settings import BaseAfeDataSettings
from organon.afe.domain.settings.base_afe_modelling_settings import BaseAfeModellingSettings
from organon.afe.domain.settings.base_afe_scoring_settings import BaseAfeScoringSettings
from organon.afe.domain.settings.base_trx_descriptor import BaseTrxDescriptor
from organon.afe.domain.settings.db_object_input import DbObjectInput
from organon.afe.domain.settings.feature_generation_settings import FeatureGenerationSettings
from organon.afe.domain.settings.record_source import RecordSource
from organon.afe.domain.settings.target_descriptor import TargetDescriptor
from organon.afe.domain.settings.temporal_grid import TemporalGrid
from organon.afe.domain.settings.trx_descriptor import TrxDescriptor
from organon.afe.domain.settings.validation_helper import check_required_attrs, raise_or_add_exception
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.core.helpers import list_helper, string_helper


class AfeSettingsValidator:
    """
    This class have static methods to validate AfeModellingSettings object before Automated Feature Extraction process
     either by raising exceptions or returning a list of exceptions
    """

    def __init__(self):
        self._columns_per_source: Dict[RecordSource, List[str]] = {}

    def validate_modelling_settings(self, modelling_settings: BaseAfeModellingSettings):
        """
        Validate AfeModellingSettings object by raising exceptions
        :param modelling_settings: AfeModellingSettings object to be validated.
        :type modelling_settings: BaseAfeModellingSettings
        """
        validate_types_modelling_settings(modelling_settings)
        self.validate_afe_modelling_settings_requirements(modelling_settings)

    def validate_scoring_settings(self, scoring_settings: BaseAfeScoringSettings):
        """
        Validate AfeModellingSettings object by raising exceptions
        :param scoring_settings: AfeModellingSettings object to be validated.
        :type scoring_settings: BaseAfeModellingSettings
        """
        validate_types_scoring_settings(scoring_settings)
        self.validate_afe_scoring_settings_requirements(scoring_settings)

    def get_modelling_validation_errors(self, modelling_settings: BaseAfeModellingSettings) -> List[Exception]:
        """
        Validate AfeModellingsSettings to return a list of exceptions that should have been raised during validation.
        :param modelling_settings: AfeModellingSettings object to be validated.
        :type modelling_settings: BaseAfeModellingSettings
        :return:
        """
        exc_list = validate_types_modelling_settings(modelling_settings, exception_list=True)
        exc_list.extend(self.validate_afe_modelling_settings_requirements(modelling_settings, exception_list=True))
        return exc_list

    @staticmethod
    def _get_horizon_list(temporal_grids: List[TemporalGrid]) -> List[int]:
        """
        Creates horizonlist from given TemporalGrid
        :param temporal_grids: List of TemporalGrid instances
        :type temporal_grids: List[TemporalGrid]
        :return: horizon list
        """
        ret_list = set()
        for temporal_grid in temporal_grids:
            for i in range(temporal_grid.length):
                ret_list.add(temporal_grid.offset + i * temporal_grid.stride)
        return list(ret_list)

    @classmethod
    def validate_afe_model_output_requirements(cls, obj: BaseAfeModelOutput, obj_str: str = "model_output",
                                               exception_list: bool = False) \
            -> List[Exception]:
        """
        Validates requirements of AfeScoringSettings object.
        :param obj: Object to be validated.
        :type obj: BaseAfeModelOutput
        :param exception_list: If true, no exceptions will be raised and a list of caught exceptions will be returned
        :type exception_list: bool
        :return: List[Exception]
        """
        exc_list = []
        required_attrs = ["output_features", "transaction_file_stats", "trx_entity_column", "trx_date_columns",
                          "target_descriptor"]
        exc_list.extend(check_required_attrs(obj, required_attrs, obj_str, exception_list=exception_list))
        return exc_list

    def validate_afe_scoring_settings_requirements(self, obj: BaseAfeScoringSettings, exception_list: bool = False) \
            -> List[Exception]:
        """
        Validates requirements of AfeScoringSettings object.
        :param obj: Object to be validated.
        :type obj: PAfeScoringSettings
        :param exception_list: If true, no exceptions will be raised and a list of caught exceptions will be returned
        :type exception_list: bool
        :return: List[Exception]
        """
        exc_list = []
        required_attrs = ["model_output", "raw_input_source"]
        exc_list.extend(check_required_attrs(obj, required_attrs, "", exception_list=exception_list))

        if obj.model_output is not None:
            exc_list.extend(self.validate_afe_model_output_requirements(obj.model_output, "model_output",
                                                                        exception_list=exception_list))

        if obj.raw_input_source is not None:
            if obj.raw_input_source.get_type() not in [RecordSourceType.TEXT, RecordSourceType.DATA_FRAME]:
                exc = KnownException("Scoring raw input source should be either a DataFrame or file path")
                raise_or_add_exception(exc, exc_list=exc_list, add_to_list=exception_list)
            exc_list.extend(self.validate_record_source_requirements(obj.raw_input_source, "raw_input_source",
                                                                     exception_list=exception_list))

        if obj.target_record_source is not None:
            if obj.target_record_source.get_type() not in [RecordSourceType.TEXT, RecordSourceType.DATA_FRAME]:
                exc = KnownException("Scoring target source should be either a DataFrame or file path")
                raise_or_add_exception(exc, exc_list=exc_list, add_to_list=exception_list)
            exc_list.extend(self.validate_record_source_requirements(obj.target_record_source, "target_record_source",
                                                                     exception_list=exception_list))

        return exc_list

    def validate_afe_modelling_settings_requirements(self, obj: BaseAfeModellingSettings,
                                                     exception_list: bool = False) -> List[Exception]:
        """
        Validates requirements of AfeModellingSettings object.
        :param obj: Object to be validated.
        :type obj: BaseAfeModellingSettings
        :param exception_list: If true, no exceptions will be raised and a list of caught exceptions will be returned
        :type exception_list: bool
        :return: List[Exception]
        """
        exc_list = []
        required_attrs = ["data_source_settings", "afe_learning_type"]
        exc_list.extend(check_required_attrs(obj, required_attrs, "", exception_list=exception_list))

        if obj.afe_learning_type is not None:
            if obj.afe_learning_type == AfeLearningType.Supervised:
                if obj.algorithm_settings is not None:
                    exc_list.extend(
                        self.validate_supervised_algorithm_settings_requirements(obj.algorithm_settings,
                                                                                 "supervised_algorithm_settings",
                                                                                 exception_list=exception_list))

            elif obj.afe_learning_type == AfeLearningType.Unsupervised:
                if obj.algorithm_settings is not None:
                    exc_list.extend(
                        self.validate_unsupervised_algorithm_settings_requirements(obj.algorithm_settings,
                                                                                   "unsupervised_algorithm_settings",
                                                                                   exception_list=exception_list))
            if obj.data_source_settings is not None:
                exc_list.extend(
                    self.validate_data_source_requirements(
                        obj.data_source_settings, "data_source_settings",
                        is_supervised=obj.afe_learning_type == AfeLearningType.Supervised,
                        exception_list=exception_list))

        if obj.output_settings is not None:
            exc_list.extend(
                self.validate_output_settings_requirements(obj.output_settings, "output_settings",
                                                           exception_list=exception_list))

        return exc_list

    def validate_data_source_requirements(self, obj: BaseAfeDataSettings, obj_str: str = "data_source_settings",
                                          is_supervised: bool = False, exception_list: bool = False) -> List[Exception]:
        """
        Validates requirements of AfeDataSettings object.
        :param is_supervised: Is the setting supervised?
        :param obj: Object to be validated.
        :type obj: BaseAfeDataSettings
        :param obj_str: A string to define the object
        :type obj_str: str
        :param exception_list: If true, no exceptions will be raised and a list of caught exceptions will be returned
        :type exception_list: bool
        :return: List[Exception]
        """
        exc_list = []
        required_attrs = ["trx_descriptor", "target_descriptor",
                          "target_record_source_list"]

        if obj.trx_descriptor is not None \
                and obj.trx_descriptor.modelling_raw_input_source is not None \
                and obj.trx_descriptor.modelling_raw_input_source.get_type() == RecordSourceType.DATABASE:
            required_attrs.append("entity_table")

        exc_list.extend(check_required_attrs(obj, required_attrs, obj_str, exception_list=exception_list))

        if obj.trx_descriptor is not None:
            exc_list.extend(
                self.validate_trx_descriptor_requirements(obj.trx_descriptor,
                                                          f"{obj_str}.trx_descriptor",
                                                          exception_list=exception_list))

        if obj.target_descriptor is not None:
            exc_list.extend(self.validate_target_descriptor_requirements(obj.target_descriptor,
                                                                         f"{obj_str}.target_descriptor",
                                                                         is_supervised,
                                                                         obj.target_record_source_list,
                                                                         exception_list=exception_list))

        if obj.target_record_source_list is not None:
            if len(obj.target_record_source_list) == 0:
                exc = KnownException(f"{obj_str}.target_record_source_list cannot be empty.")
                raise_or_add_exception(exc, exc_list, add_to_list=exception_list)
            else:
                for i, record_source in enumerate(obj.target_record_source_list):
                    exc_list.extend(self.validate_record_source_requirements(
                        record_source,
                        f"{obj_str}.target_record_source_list[{i}]",
                        exception_list=exception_list))

        return exc_list

    def validate_trx_descriptor_requirements(self, obj: BaseTrxDescriptor,
                                             obj_str: str = "trx_descriptor",
                                             exception_list: bool = False) -> List[Exception]:
        """
        Validates requirements of TransactionFileDesciptor object.
        :param obj: Object to be validated.
        :type obj: BaseTrxDescriptor
        :param obj_str: A string to define the object
        :type obj_str: str
        :param exception_list: If true, no exceptions will be raised and a list of caught exceptions will be returned
        :type exception_list: bool
        :return: List[Exception]
        """
        exc_list = []

        if not isinstance(obj, TrxDescriptor):
            raise ValueError

        required_attrs = ["modelling_raw_input_source", "entity_column_name", "feature_gen_setting"]
        feature_generation_setting = obj.feature_gen_setting

        exc_list.extend(check_required_attrs(obj, required_attrs, obj_str, exception_list=exception_list))

        if not string_helper.is_null_or_empty(obj.entity_column_name):
            exc_list.extend(self._check_column_in_source(obj.entity_column_name, obj.modelling_raw_input_source,
                                                         "transaction source",
                                                         exception_list=exception_list))

        if obj.modelling_raw_input_source is not None:
            mod_raw_inp_src_exc = self.validate_record_source_requirements(obj.modelling_raw_input_source,
                                                                           f"{obj_str}.modelling_raw_input_source",
                                                                           exception_list=exception_list)

            exc_list.extend(mod_raw_inp_src_exc)

        if feature_generation_setting is not None:
            exc_list.extend(
                self.validate_afe_feature_generation_settings_requirements(
                    feature_generation_setting,
                    f"{obj_str}.feature_gen_setting",
                    raw_input_source=obj.modelling_raw_input_source,
                    exception_list=exception_list))

        return exc_list

    @classmethod
    def validate_record_source_requirements(cls, obj: RecordSource, obj_str: str = "record_source",
                                            exception_list: bool = False) -> List[Exception]:
        """
        Validates requirements of RecordSource object.
        :param obj: Object to be validated.
        :type obj: RecordSource
        :param obj_str: A string to define the object
        :type obj_str: str
        :param exception_list: If true, no exceptions will be raised and a list of caught exceptions will be returned
        :type exception_list: bool
        :return: List[Exception]
        """
        exc_list = []
        required_attrs = ["source"]
        exc_list.extend(check_required_attrs(obj, required_attrs, obj_str, exception_list))
        if obj.get_type() == RecordSourceType.DATABASE and obj.source is not None:
            exc_list.extend(
                cls.validate_db_object_input_requirements(obj.source, f"{obj_str}.source", exception_list))
        return exc_list

    def validate_target_descriptor_requirements(self, obj: TargetDescriptor,
                                                obj_str: str = "target_descriptor",
                                                is_supervised: bool = False,
                                                target_record_source_list: List[RecordSource] = None,
                                                exception_list: bool = False) -> List[Exception]:
        """
        Validates requirements of TargetFileDescriptor object.
        :param is_supervised: If algorithm supervised then target_column is needed
        :param obj: Object to be validated.
        :type obj: TargetDescriptor
        :param obj_str: A string to define the object
        :type obj_str: str
        :param exception_list: If true, no exceptions will be raised and a list of caught exceptions will be returned
        :type exception_list: bool
        :return: List[Exception]
        """
        exc_list = []
        required_attrs = ["entity_column_name"]
        if is_supervised:
            required_attrs.append("target_column")

        columns_to_check = []
        required_attr_exceptions = check_required_attrs(obj, required_attrs, obj_str, exception_list)
        exc_list.extend(required_attr_exceptions)
        self._add_if_not_null_or_empty(obj.entity_column_name, columns_to_check)
        if obj.date_column is not None:
            date_col_exc = self.validate_afe_date_column_requirements(obj.date_column, f"{obj_str}.date_column",
                                                                      exception_list=exception_list)
            exc_list.extend(date_col_exc)
            self._add_if_not_null_or_empty(obj.date_column.column_name, columns_to_check)
            exc_list.extend(
                self._validate_date_column_for_target_source_list(obj.date_column, target_record_source_list,
                                                                  f"{obj_str}.date_column", exception_list))

        if is_supervised and obj.target_column is not None:
            target_col_exc = self.validate_afe_target_column_requirements(obj.target_column,
                                                                          f"{obj_str}.target_column",
                                                                          exception_list=exception_list)
            exc_list.extend(target_col_exc)
            self._add_if_not_null_or_empty(obj.target_column.column_name, columns_to_check)

        if not is_supervised and obj.target_column is not None:
            exc = KnownException(f"{obj_str}.target_column should be None for unsupervised modelling.")
            raise_or_add_exception(exc, exc_list, add_to_list=exception_list)

        if target_record_source_list is not None:
            for i, source in enumerate(target_record_source_list):
                for col in columns_to_check:
                    exc_list.extend(self._check_column_in_source(col, source, f"target_record_source_list[{i}]"))

        return exc_list

    def validate_afe_feature_generation_settings_requirements(self, obj: FeatureGenerationSettings,
                                                              obj_str: str = "feature_generation_settings",
                                                              raw_input_source: RecordSource = None,
                                                              exception_list: bool = False):
        """
        Validates requirements of AfeDateColumnSettings object.
        :param obj: Object to be validated.
        :type obj: FeatureGenerationSettings
        :param obj_str: A string to define the object
        :type obj_str: str
        :param exception_list: If true, no exceptions will be raised and a list of caught exceptions will be returned
        :type exception_list: bool
        :return: List[Exception]
        """
        exc_list = []
        columns_to_check = []
        if obj.date_column is not None:
            date_col_excs = check_required_attrs(obj,
                                                 ["date_resolution", "temporal_grids"],
                                                 obj_str, exception_list)
            exc_list.extend(date_col_excs)

            date_col_exc = self.validate_afe_date_column_requirements(obj.date_column, f"{obj_str}.date_column",
                                                                      exception_list=exception_list)
            exc_list.extend(date_col_exc)
            exc_list.extend(self._validate_date_column_for_raw_input_source(obj.date_column, raw_input_source,
                                                                            f"{obj_str}.date_column",
                                                                            exception_list=exception_list))
            self._add_if_not_null_or_empty(obj.date_column.column_name, columns_to_check)
        else:
            obj.temporal_grids = [TemporalGrid(stride=1, length=1, offset=1)]
            obj.date_resolution = DateResolution.Day

        if not list_helper.is_null_or_empty(obj.temporal_grids):
            temp_grid_excs = []
            for i, temp_grid in enumerate(obj.temporal_grids):
                temp_grid_excs.extend(self.validate_temporal_grid_requirements(temp_grid,
                                                                               f"{obj_str}.temporal_grids[{i}]",
                                                                               exception_list=exception_list))
            obj.horizon_list = self._get_horizon_list(obj.temporal_grids) if len(temp_grid_excs) == 0 else None
            exc_list.extend(temp_grid_excs)
        else:
            exc = KnownException(f"{obj_str}.temporal_grids cannot be empty")
            raise_or_add_exception(exc, exc_list, add_to_list=exception_list)

        if not list_helper.is_null_or_empty(obj.dimension_columns):
            columns_to_check.extend([col for col in obj.dimension_columns
                                     if col != AfeStaticObjects.empty_dimension_column])
        if not list_helper.is_null_or_empty(obj.quantity_columns):
            columns_to_check.extend([col for col in obj.quantity_columns
                                     if col != AfeStaticObjects.empty_quantity_column])

        for col in columns_to_check:
            exc_list.extend(self._check_column_in_source(col, raw_input_source,
                                                         "transaction source",
                                                         exception_list=exception_list))

        return exc_list

    @classmethod
    def validate_afe_date_column_requirements(cls, obj: AfeDateColumn, obj_str: str = "date_column",
                                              exception_list: bool = False):
        """
        Validates requirements of AfeDateColumn object.

        :param obj: Object to be validated.
        :type obj: AfeDateColumn
        :param obj_str: A string to define the object
        :type obj_str: str
        :param exception_list: If true, no exceptions will be raised and a list of caught exceptions will be returned
        :type exception_list: bool
        :return: List[Exception]
        """
        exc_list = []
        date_col_excs = check_required_attrs(obj,
                                             ["column_name", "date_column_type"],
                                             obj_str, exception_list)
        exc_list.extend(date_col_excs)
        if obj.date_column_type == AfeDateColumnType.CustomFormat:
            if obj.custom_format is None:
                exc = KnownException(f"Please provide {obj_str}.custom_format.")
                raise_or_add_exception(exc, exc_list, add_to_list=exception_list)

        return exc_list

    @classmethod
    def validate_afe_target_column_requirements(cls, obj: AfeTargetColumn, obj_str: str = "target_column",
                                                exception_list: bool = False) -> List[Exception]:
        """
        Validates requirements of AfeTargetColumn object.
        :param obj: Object to be validated.
        :type obj: AfeTargetColumn
        :param obj_str: A string to define the object
        :type obj_str: str
        :param exception_list: If true, no exceptions will be raised and a list of caught exceptions will be returned
        :type exception_list: bool
        :return: List[Exception]
        """
        exc_list = []
        required_attrs = ["column_name", "target_column_type"]
        target_col_exc = check_required_attrs(obj, required_attrs,
                                              obj_str, exception_list)
        exc_list.extend(target_col_exc)

        if obj.target_column_type == AfeTargetColumnType.Binary:
            binary_target_info = obj.binary_target_info
            if binary_target_info is not None:
                bin_target_fields = [binary_target_info.positive_category, binary_target_info.negative_category,
                                     binary_target_info.indeterminate_category, binary_target_info.exclusion_category]

                if bin_target_fields[0] is None or bin_target_fields[1] is None or \
                        sum(x is not None for x in bin_target_fields) == 3:
                    exc = KnownException(f"{obj_str}.binary_target_info is invalid.")
                    raise_or_add_exception(exc, exc_list, add_to_list=exception_list)
            else:
                exc = KnownException(f"{obj_str}.binary_target_info should be provided.")
                raise_or_add_exception(exc, exc_list, add_to_list=exception_list)

        return exc_list

    @classmethod
    def validate_supervised_algorithm_settings_requirements(cls, obj: AfeSupervisedAlgorithmSettings,
                                                            obj_str: str = "supervised_algorithm_settings",
                                                            exception_list: bool = False) -> List[Exception]:
        """
        Validates requirements of AfeSupervisedAlgorithmSettings object.
        :param obj: Object to be validated.
        :type obj: AfeSupervisedAlgorithmSettings
        :param obj_str: A string to define the object
        :type obj_str: str
        :param exception_list: If true, no exceptions will be raised and a list of caught exceptions will be returned
        :type exception_list: bool
        :return: List[Exception]
        """
        exc_list = []
        exc_list.extend(cls.validate_afe_algorithm_settings_requirements(obj, obj_str, exception_list=exception_list))

        return exc_list

    @classmethod
    def validate_unsupervised_algorithm_settings_requirements(cls, obj: AfeUnsupervisedAlgorithmSettings,
                                                              obj_str: str = "unsupervised_algorithm_settings",
                                                              exception_list: bool = False) -> List[Exception]:
        """
        Validates requirements of AfeUnsupervisedAlgorithmSettings object.
        :param obj: Object to be validated.
        :type obj: AfeUnsupervisedAlgorithmSettings
        :param obj_str: A string to define the object
        :type obj_str: str
        :param exception_list: If true, no exceptions will be raised and a list of caught exceptions will be returned
        :type exception_list: bool
        :return: List[Exception]
        """
        exc_list = []
        exc_list.extend(cls.validate_afe_algorithm_settings_requirements(obj, obj_str, exception_list=exception_list))
        return exc_list

    @classmethod
    def validate_output_settings_requirements(cls, obj: AfeOutputSettings, obj_str: str = "output_settings",
                                              exception_list: bool = False) -> List[Exception]:
        """
        Validates requirements of AfeOutputSettings object.
        :param obj: Object to be validated.
        :type obj: AfeOutputSettings
        :param obj_str: A string to define the object
        :type obj_str: str
        :param exception_list: If true, no exceptions will be raised and a list of caught exceptions will be returned
        :type exception_list: bool
        :return: List[Exception]
        """
        exc_list = []
        output_params = [obj.enable_all_feature_lookup_output_to_csv, obj.enable_feature_lookup_output_to_csv,
                         obj.enable_write_output]
        prefix_needed = sum([1 for val in output_params if val is True]) > 0
        if prefix_needed:
            if string_helper.is_null_or_empty(obj.output_prefix):
                exc = KnownException(f"Please give {obj_str}.output_prefix which will be used as a prefix in "
                                     f"output file names."
                                     "If you do not want your output files of a previous AFE execution to be "
                                     "overwritten, either output_folder or output_prefix should be "
                                     "given a unique value")
                raise_or_add_exception(exc, exc_list, exception_list)

        return exc_list

    @classmethod
    def validate_afe_algorithm_settings_requirements(cls, obj: AfeAlgorithmSettings,
                                                     obj_str: str = "algorithm_settings",
                                                     exception_list: bool = False) -> List[Exception]:
        """
        Validates requirements of AfeAlgorithmSettings object.
        :param obj: Object to be validated.
        :type obj: AfeAlgorithmSettings
        :param obj_str: A string to define the object
        :type obj_str: str
        :param exception_list: If true, no exceptions will be raised and a list of caught exceptions will be returned
        :type exception_list: bool
        :return: List[Exception]
        """
        # pylint: disable=unused-argument
        exc_list = []

        return exc_list

    @classmethod
    def validate_db_object_input_requirements(cls, obj: DbObjectInput, obj_str: str = "DbObjectInput",
                                              exception_list: bool = False) -> List[Exception]:
        """
        Validates requirements of DbObjectDto object.
        :param obj: Object to be validated.
        :type obj: DbObjectInput
        :param obj_str: A string to define the object
        :type obj_str: str
        :param exception_list: If true, no exceptions will be raised and a list of caught exceptions will be returned
        :type exception_list: bool
        :return: List[Exception]
        """
        exc_list = []
        required_attrs = ["connection_name"]
        exc_list.extend(check_required_attrs(obj, required_attrs, obj_str, exception_list=exception_list))
        if obj.sql_statement is None and (obj.schema_name is None or obj.table_name is None):
            exc = KnownException(
                f"Either give {obj_str}.sql_statement or {obj_str}.schema_name with {obj_str}.table_name")
            raise_or_add_exception(exc, exc_list, add_to_list=exception_list)
        return exc_list

    @classmethod
    def validate_temporal_grid_requirements(cls, obj: TemporalGrid, obj_str: str = "TemporalGrid",
                                            exception_list: bool = False) -> List[Exception]:
        """
        Validates requirements of TemporalGrid object.
        :param TemporalGrid obj: Object to be validated.
        :param str obj_str: A string to define the object
        :param bool exception_list: If true,
            no exceptions will be raised and a list of caught exceptions will be returned
        :return: List[Exception]
        """
        exc_list = []
        required_attrs = ["length", "stride", "offset"]
        exc_list.extend(check_required_attrs(obj, required_attrs, obj_str, exception_list=exception_list))
        if obj.offset <= 0:
            exc = KnownException(f"{obj_str}.offset should be a positive integer")
            raise_or_add_exception(exc, exc_list, add_to_list=exception_list)
        if obj.length <= 0:
            exc = KnownException(f"{obj_str}.length should be a positive integer")
            raise_or_add_exception(exc, exc_list, add_to_list=exception_list)
        if obj.stride <= 0:
            exc = KnownException(f"{obj_str}.stride should be a positive integer")
            raise_or_add_exception(exc, exc_list, add_to_list=exception_list)
        return exc_list

    @classmethod
    def _validate_date_column_for_raw_input_source(cls, date_column: AfeDateColumn, raw_input_source: RecordSource,
                                                   obj_str: str,
                                                   exception_list: bool = False):
        # pylint: disable=unused-argument
        return []

    @classmethod
    def _validate_date_column_for_target_source_list(cls, date_column: AfeDateColumn,
                                                     target_record_source_list: List[RecordSource],
                                                     obj_str: str,
                                                     exception_list: bool = False):
        # pylint: disable=unused-argument
        return []

    def _check_column_in_source(self, column: str, source: RecordSource, source_name: str,
                                exception_list: bool = False):
        exc_list = []
        if source not in self._columns_per_source:
            self._columns_per_source[source] = self._get_columns_from_source(source)
        if column not in self._columns_per_source[source]:
            exc = KnownException(f"Column '{column}' not found in {source_name}")
            raise_or_add_exception(exc, exc_list, add_to_list=exception_list)
        return exc_list

    @classmethod
    def _get_columns_from_source(cls, source: RecordSource):
        if source.get_type() == RecordSourceType.DATA_FRAME:
            return list(source.source.columns)
        if source.get_type() == RecordSourceType.TEXT:
            return list(pd.read_csv(source.source, sep=",", nrows=0).columns)
        raise NotImplementedError

    @staticmethod
    def _add_if_not_null_or_empty(string_val: str, _list: List[str]):
        if not string_helper.is_null_or_empty(string_val):
            _list.append(string_val)



""" This module includes PredictorFactory class """
from organon.idq.domain.controls.helpers.algorithms.random_forest_predictor import RandomForestPredictor
from organon.idq.domain.enums.dq_pred_algorithm_type import DqPredAlgorithmType


class PredictorFactory:
    """ Class for PredictorFactory"""

    @staticmethod
    def get_predictor(predictor: DqPredAlgorithmType):
        """ :returns predictor instance according to its type """
        if predictor == DqPredAlgorithmType.RANDOM_FOREST_REG:
            return RandomForestPredictor()
        raise NotImplementedError



"""Includes XGBoostClassifier class."""
from xgboost import XGBClassifier

from organon.ml.modelling.algorithms.core.abstractions.base_sklearn_classifier import BaseSklearnClassifier


class XGBoostClassifier(BaseSklearnClassifier[XGBClassifier]):
    """Wrapper for XGBClassifier"""

    @property
    def classifier_class(self):
        return XGBClassifier



"""Includes DqComparisonResult class"""
from typing import Optional, Union

from organon.idq.core.enums.data_entity_type import DataEntityType
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode
from organon.idq.core.helpers.formatting_helper import FormattingHelper
from organon.idq.domain.enums.dq_test_group_type import DqTestGroupType


class DqComparisonResult:
    """todo"""

    def __init__(self, data_entity: DataEntityType, data_entity_name: str,
                 test_group: DqTestGroupType, result_code: DqComparisonResultCode = None,
                 message: str = None, property_code: str = None
                 ):
        self.data_entity = data_entity
        self.data_entity_name = data_entity_name
        self.test_group = test_group
        self.result_code = result_code
        self.message = message
        self.property_code = property_code
        self.property_key_numeric: Optional[float] = None
        self.property_key_nominal: Optional[str] = None
        self.property_value_numeric: Optional[float] = None
        self.property_value_nominal: Optional[str] = None
        self.signal_type = None

    def set_property_key(self, key: Union[str, float]):
        """Sets property_key_nominal/property_key_numeric value"""
        if isinstance(key, str):
            self.property_key_nominal = key
        elif isinstance(key, float):
            self.property_key_numeric = FormattingHelper.to_nullable(key)
        else:
            raise ValueError("key should be either str or float")

    def set_property_key_value_pair(self, key: Union[str, float], value: Union[str, float]):
        """Sets property_key-propertyvalue pair"""
        self.set_property_key(key)
        if isinstance(value, str):
            self.property_value_nominal = value
        elif isinstance(value, float):
            self.property_value_numeric = FormattingHelper.to_nullable(value)
        else:
            raise ValueError("Value should be either str or float")



"""
This module includes AfeDateColumn class.
"""
from organon.afe.domain.common.reader_helper import get_values_from_kwargs
from organon.afe.domain.enums.afe_date_column_type import AfeDateColumnType


class AfeDateColumn:
    """
    Class for columns of type date
    """

    ATTR_DICT = {
        "column_name": str,
        "date_column_type": AfeDateColumnType,
        "custom_format": str,
        "db_custom_format": str
    }

    def __init__(self, **kwargs):
        self.column_name: str = None
        self.date_column_type: AfeDateColumnType = None
        self.custom_format: str = None
        self.db_custom_format: str = None

        get_values_from_kwargs(self, AfeDateColumn.ATTR_DICT, kwargs)

    def __deepcopy__(self, memo):
        output = AfeDateColumn()
        output.column_name = self.column_name
        output.date_column_type = self.date_column_type
        output.custom_format = self.custom_format
        output.db_custom_format = self.db_custom_format
        return output



"""Includes HPOOutput class."""
from typing import Dict, Any

import pandas as pd

from organon.ml.modelling.algorithms.core.abstractions.base_modeller import BaseModeller


class HPOOutput:
    """Output of HPOService"""

    def __init__(self):
        self.best_modeller: BaseModeller = None
        self.best_params: Dict[str, Any] = None
        self.summary: pd.DataFrame = None



"""Includes PartitionInfo class."""
from typing import Union, List

from organon.fl.core.helpers import list_helper
from organon.idq.domain.settings.date_value_definition import DateValueDefinition


class PartitionInfo:
    """Dq column partition info."""

    def __init__(self):
        self.column_name: str = None
        self.column_values: Union[List[DateValueDefinition], List[float], List[str]] = None

    def to_str(self):
        """return partition info as str"""
        return f"[{self.column_name}={self.get_col_val_str()}]"

    def get_col_val_str(self) -> str:
        """returns partition info column values as str"""
        column_values = self.column_values
        if list_helper.is_null_or_empty(column_values):
            return ""
        col_val_str_parts = []
        for val in column_values:
            if isinstance(val, DateValueDefinition):
                col_val_str_parts.append(self.__get_col_val_str_for_date_value_partition_definition(val))
            else:
                col_val_str_parts.append(str(val))

        col_val_str = "(" + ",".join(col_val_str_parts) + ")"
        return col_val_str

    @classmethod
    def __get_col_val_str_for_date_value_partition_definition(cls, definition: DateValueDefinition) -> str:
        parts = []
        if definition.year is not None:
            parts.append(f"y={definition.year}")
        if definition.month is not None:
            parts.append(f"m={definition.month}")
        if definition.day is not None:
            parts.append(f"d={definition.day}")
        if definition.hour is not None:
            parts.append(f"h={definition.hour}")
        ret = "{" + ",".join(parts) + "}"
        return ret



"""Includes DataPartition class."""
import random
from typing import List

import numpy as np

from organon.fl.core.businessobjects.idataframe import IDataFrame
from organon.fl.core.businessobjects.idata_partition import IDataPartition
from organon.fl.mathematics.businessobjects.tuple_generic_collection import TupleGenericCollection

epsilon_val = np.finfo(float).eps


class DataPartition(IDataPartition):
    """This class helps with partitioning a dataframe according to training settings."""

    def __init__(self, frame: IDataFrame = None):
        super().__init__()
        self.__frame: IDataFrame = frame
        self.__training_ratio: float = 1.0
        self.__training_indices: List[int] = None
        self.__validation_indices: List[int] = None
        self.__strata_info: TupleGenericCollection[int] = None

    @property
    def frame(self):
        return self.__frame

    @property
    def training_ratio(self):
        return self.__training_ratio

    @property
    def validation_ratio(self):
        """Returns ratio of validation (1.0 - training ratio)"""
        return 1.0 - self.__training_ratio

    @property
    def strata_info(self):
        return self.__strata_info

    @property
    def training_indices(self) -> List[int]:
        return self.__training_indices

    @training_indices.setter
    def training_indices(self, value):
        self.__training_indices = value

    @property
    def validation_indices(self) -> List[int]:
        return self.__validation_indices

    @validation_indices.setter
    def validation_indices(self, value):
        self.validation_indices = value

    def partition(self, training_ratio: float = 1.0):
        system_random = random.SystemRandom()
        rows = self.__frame.row_count
        if epsilon_val < training_ratio <= 1.0:
            self.__training_ratio = training_ratio
            if self.__training_ratio >= 1.0:
                self.__training_indices = list(range(rows))
            else:
                t_list = []
                v_list = []
                for i in range(rows):
                    if system_random.uniform(0, 1) <= self.__training_ratio:
                        t_list.append(i)
                    else:
                        v_list.append(i)
                self.__training_indices = t_list
                self.__validation_indices = v_list
        else:
            self.__training_ratio = 1.0
            self.__training_indices = list(range(rows))

    def set_frame(self, frame: IDataFrame):
        """Changes frame of the partition"""
        self.__frame = frame



"""
This module includes AfeModellingSettings class.
"""

from organon.afe.domain.settings.afe_data_settings import AfeDataSettings
from organon.afe.domain.settings.afe_output_settings import AfeOutputSettings
from organon.afe.domain.settings.base_afe_modelling_settings import BaseAfeModellingSettings


class AfeModellingSettings(BaseAfeModellingSettings[AfeDataSettings, AfeOutputSettings]):
    """
    Modelling Settings for Automated Feature Extraction
    """
    ATTR_DICT = {
        "data_source_settings": AfeDataSettings,
        "output_settings": AfeOutputSettings,
    }
    ATTR_DICT.update(BaseAfeModellingSettings.ATTR_DICT)



"""Includes BaseModeller class."""
import abc
from typing import Any, Dict, Union

import numpy as np
import pandas as pd
from sklearn import clone

# THIS CLASS IS THE BASE CLASS FOR OUR REGRESSORS AND CLASSIFIERS. ALL OF THESE REGRESSORS AND CLASSIFIERS SHOULD BE
# USABLE IN HYPERPARAMETER OPTIMIZATION MODULE. THAT MODULE REQUIRES SOME MODULES TO BE DEFINED IN
# ESTIMATOR(MODELLER AS WE CALL IT) CLASSES AND THERE ARE SOME REQUIREMENTS ON __init__ METHOD AND PARAMETERS TOO.
# THESE REQUIREMENTS EXIST BECAUSE HYPERPARAMETER OPTIMIZATION MODULE USES PIPELINE, GRIDSEARCHCV etc. FROM SKLEARN
# PACKAGE AND OUR CLASSIFIERS AND REGRESSORS SHOULD BE ABLE TO BE USED WITH THESE.
# CHECK REQUIREMENTS FOR SKLEARN ESTIMATORS IN https://scikit-learn.org/stable/developers/develop.html
from organon.ml.modelling.algorithms.core.enums.modeller_type import ModellerType


class BaseModeller(metaclass=abc.ABCMeta):
    """Base class for all modellers"""

    # REQUIRED FOR SKLEARN GridSearch. Set to classifier, regressor etc. in child classes
    # https://scikit-learn.org/stable/developers/develop.html#estimator-types
    _estimator_type = None

    def __init__(self, **params):
        # See __init__ method requirements in https://scikit-learn.org/stable/developers/develop.html#instantiation
        # DO NOT OVERRIDE __INIT__ IN CHILD CLASSES. THIS __init__ IN THIS BASE CLASS IS WRITTEN IN A WAY THAT NEW
        # MODELLER CLASSES DO NOT HAVE TO DECLARE AN __init__ WHICH WILL ABIDE SKLEARN RULES. YOU SHOULD ONLY IMPLEMENT
        # _get_params_with_defaults METHOD WHICH WILL RETURN A DICTIONARY OF MODELLER PARAMETERS AND
        # THEIR DEFAULT VALUES. __init__ IMPLEMENTATION IN THIS CLASS WILL HANDLE FURTHER REQUIREMENTS

        self.set_params(**self._get_params_with_defaults())
        self.set_params(**params)

    def fit(self, train_data: Union[pd.DataFrame, np.ndarray], target_data: Union[pd.DataFrame, pd.Series, np.ndarray],
            **kwargs):
        """Fits model and returns self"""
        # THIS METHOD IS REQUIRED FOR SKLEARN ESTIMATORS. DO NOT CHANGE ITS SIGNATURE AND DO NOT DIRECTLY OVERRIDE THIS
        # TO IMPLEMENT FITTING. OVERRIDE _fit ABSTRACT METHOD INSTEAD.
        if isinstance(train_data, np.ndarray):
            train_data = pd.DataFrame(train_data)
        if isinstance(target_data, np.ndarray):
            target_data = pd.Series(target_data)
        self.train_data_columns_ = train_data.columns.tolist()  # pylint: disable=attribute-defined-outside-init
        self._fit(train_data, target_data, **kwargs)
        self.fitted_ = True  # pylint: disable=attribute-defined-outside-init
        return self

    @abc.abstractmethod
    def _fit(self, train_data: pd.DataFrame, target_data: Union[pd.DataFrame, pd.Series], **kwargs):
        """Fits the model"""

    def predict(self, data: Union[pd.DataFrame, np.ndarray]) -> pd.DataFrame:
        """Predicts using fitted model and returns predictions"""
        # THIS METHOD IS REQUIRED FOR SKLEARN ESTIMATORS. DO NOT CHANGE ITS SIGNATURE AND DO NOT DIRECTLY OVERRIDE THIS
        # TO IMPLEMENT FITTING. OVERRIDE _predict ABSTRACT METHOD INSTEAD.
        self._check_is_fitted()
        if isinstance(data, np.ndarray):
            data = pd.DataFrame(data)
        return self._predict(data)

    def is_fitted(self) -> bool:
        """Returns True if model is fitted"""
        return hasattr(self, "fitted_") and self.fitted_

    def _check_is_fitted(self):
        """Raise error if modeller is not fitted yet"""
        if not self.is_fitted():
            raise ValueError("Modeller not fitted yet")

    @abc.abstractmethod
    def _predict(self, data: pd.DataFrame) -> pd.DataFrame:
        """Prediction using fitted model"""

    def score(self, train_data: pd.DataFrame, target_data: Union[pd.DataFrame, pd.Series],
              sample_weight=None) -> float:
        """Returns a score value after predicting train_data and comparing it with actual target_data"""
        # THIS METHOD IS REQUIRED FOR SKLEARN GridSearchCV ETC.
        # DO NOT CHANGE ITS SIGNATURE AND DO NOT DIRECTLY OVERRIDE THIS
        # TO IMPLEMENT FITTING. OVERRIDE _score ABSTRACT METHOD INSTEAD.
        self._check_is_fitted()
        return self._score(train_data, target_data, sample_weight=sample_weight)

    @abc.abstractmethod
    def _score(self, train_data: pd.DataFrame, target_data: Union[pd.DataFrame, pd.Series],
               sample_weight=None) -> float:
        """Returns a score value after predicting train_data and comparing it with actual target_data"""

    @abc.abstractmethod
    def _get_params_with_defaults(self) -> Dict[str, Any]:
        """Returns modeller parameter names and their default values"""

    def get_params(self, deep=True):
        """Returns dictionary of parameters and their values"""
        # pylint: disable=unused-argument
        # THIS METHOD IS REQUIRED FOR SKLEARN ESTIMATORS
        # https://scikit-learn.org/stable/developers/develop.html#get-params-and-set-params
        out = {}
        for key in self._get_params_with_defaults():
            value = getattr(self, key)
            out[key] = value
        return out

    def set_params(self, **params):
        """Set modeller parameter values"""
        # THIS METHOD IS REQUIRED FOR SKLEARN ESTIMATORS
        # https://scikit-learn.org/stable/developers/develop.html#get-params-and-set-params
        if not params:
            return self
        valid_params = self._get_params_with_defaults()
        for param, value in params.items():
            if param not in valid_params:
                raise ValueError(
                    f"Invalid parameter {param} for estimator {self}. "
                    "Check the list of available parameters "
                    "with `estimator.get_params().keys()`."
                )
            setattr(self, param, value)
        return self

    def clone(self):
        """Construct and return a new unfitted estimator with the same parameters."""
        return clone(self)

    @property
    @abc.abstractmethod
    def modeller_type(self) -> ModellerType:
        """Returns type of modeller"""



"""Includes SamplingSettings class."""
from dataclasses import dataclass
from typing import List

import pandas as pd

from organon.ml.common.enums.target_type import TargetType
from organon.ml.sampling.settings.enums.sampling_strategy import SamplingStrategy


@dataclass
class SamplingSettings:
    """Settings for Sampling service"""
    data: pd.DataFrame
    target_column_name: str
    target_type: TargetType
    strata_columns: List[str]
    split_test_data: bool
    test_split_ratio: float
    sampling_strategy: SamplingStrategy
    sampling_ratio: float
    data_sample_ratio: float = None



"""
This module includes Pair class.
"""
from typing import TypeVar, Generic, NamedTuple
S1 = TypeVar("S1")
T1 = TypeVar("T1")


class Pair(Generic[S1, T1], NamedTuple("Pair", [('first', S1), ('second', T1)])):
    """
    Class for defining the comparable and generic pair type.
    """
    def __eq__(self, other: "Pair"):
        return self.first == other.first and self.second == other.second

    def __lt__(self, other: "Pair"):
        return self.first < other.first

    def __le__(self, other: "Pair"):
        return self.first <= other.first

    def __gt__(self, other: "Pair"):
        return self.first > other.first

    def __ge__(self, other: "Pair"):
        return self.first >= other.first

    def __str__(self):
        return f"({self.first}, {self.second})"

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash((self.first, self.second))



"""Includes NullFeatureReductionSettings class."""
from typing import List, Optional
from pandas import DataFrame
from organon.ml.feature_reduction.settings.objects.base_feature_reduction_settings import \
    BaseFeatureReductionSettings


class NullFeatureReductionSettings(BaseFeatureReductionSettings):
    """Settings for null feature reduction"""

    def __init__(self, data: DataFrame, null_ratio_threshold: float = None,
    excluded_columns: Optional[List[str]] = None):
        super().__init__(data, excluded_columns)
        self.null_ratio_threshold = null_ratio_threshold



"""
This module includes AutoColumnDeciderSettings class.
"""
from typing import List

from organon.afe.domain.common.reader_helper import get_values_from_kwargs


class AutoColumnDeciderSettings:
    """
    Class for parameters about the AutoColumnDeciderService
    """
    ATTR_DICT = {
        "sampling_ratio": float,
        "numeric_to_dimension": int,
        "dimension_distinct_cut_off": float,
        "use_dimension_columns": bool,
        "rejected_dimension_columns": List[str],
        "use_quantity_columns": bool,
        "rejected_quantity_columns": List[str]
    }

    def __init__(self, **kwargs):
        self.sampling_ratio: float = None
        self.numeric_to_dimension: int = None
        self.dimension_distinct_cut_off: float = None
        self.use_dimension_columns: bool = None
        self.rejected_dimension_columns: List[str] = None
        self.use_quantity_columns: bool = None
        self.rejected_quantity_columns: List[str] = None

        get_values_from_kwargs(self, self.ATTR_DICT, kwargs)



"""Includes StackingEnsembleClassifier class"""
from typing import Dict, Any

from sklearn.ensemble import StackingClassifier

from organon.ml.modelling.algorithms.core.abstractions.base_sklearn_classifier import BaseSklearnClassifier


class StackingEnsembleClassifier(BaseSklearnClassifier):
    """Wrapper for sklearn StackingClassifier"""

    @property
    def classifier_class(self):
        return StackingClassifier

    def _get_params_with_defaults(self) -> Dict[str, Any]:
        return StackingClassifier(estimators=[]).get_params()



"""Module for IMatrix"""
from abc import ABCMeta, abstractmethod
import numpy as np


class IMatrix(metaclass=ABCMeta):
    """Matrix interface for sweep algorithm"""
    @property
    @abstractmethod
    def shape(self) -> tuple:
        """Shape of the matrix"""

    @property
    @abstractmethod
    def is_centralized(self) -> bool:
        """Get the centralized information"""

    @property
    @abstractmethod
    def data(self) -> np.ndarray:
        """Get the data"""

    @property
    @abstractmethod
    def means(self) -> np.ndarray:
        """Get column means of the matrix"""

    @abstractmethod
    def get_column(self, k: int) -> np.ndarray:
        """Get specified column of the matrix"""

    @abstractmethod
    def mul_vector(self, vector: np.ndarray) -> np.ndarray:
        """Calculates Matrix Vector product"""

    @abstractmethod
    def centralize_columns(self):
        """Centralize all columns"""



"""Includes RFSelectionOutput class"""
from organon.ml.feature_selection.domain.objects.base_selection_output import BaseSelectionOutput


class RFSelectionOutput(BaseSelectionOutput):
    """Output of random forest selection"""



"""Includes StopWatch class."""
import time

from organon.fl.core.helpers import process_info_helper


class StopWatch:
    """Class for tracking execution time between lines."""

    def __init__(self, start: bool = False, with_cpu_times=False):
        self.initial_time = None
        self.with_cpu_times = with_cpu_times
        self.initial_cpu_times = None
        if start:
            self.start()

    def start(self):
        """Starts the timer."""
        self.initial_time = time.time()
        if self.with_cpu_times:
            self.initial_cpu_times = process_info_helper.get_cpu_times()

    def get_elapsed_seconds(self, restart: bool = False):
        """Returns the time(in seconds) passed since last re/start."""
        retval = time.time() - self.initial_time
        if restart:
            self.restart()
        return retval

    def get_cpu_times(self, restart: bool = False):
        """Gets cpu times used since start."""
        if not self.with_cpu_times:
            return None
        curr_cpu_times = process_info_helper.get_cpu_times()
        if restart:
            self.restart()
        return curr_cpu_times[0] - self.initial_cpu_times[0], curr_cpu_times[1] - self.initial_cpu_times[1]

    def restart(self):
        """Restarts timer."""
        self.initial_time = time.time()
        if self.with_cpu_times:
            self.initial_cpu_times = process_info_helper.get_cpu_times()



"""Includes BaseDqReportHelper class"""
import abc
from typing import List, Dict

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.helpers import list_helper, string_helper
from organon.idq.core.dq_constants import DqConstants
from organon.idq.core.enums.data_entity_type import DataEntityType
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.reporting.objects.base_dq_output_report import BaseDqOutputReport
from organon.idq.domain.reporting.objects.base_dq_report_helper_params import BaseDqReportHelperParams
from organon.idq.domain.reporting.objects.dq_comparison_sample_report_input import DqComparisonSampleReportInput
from organon.idq.domain.reporting.objects.signal_details import SignalDetails
from organon.idq.domain.settings.partition_info import PartitionInfo


class BaseDqReportHelper(metaclass=abc.ABCMeta):
    """Class for base dq report  helper"""

    def __init__(self, base_dq_output_helper_params: BaseDqReportHelperParams):
        self.data_source_name = base_dq_output_helper_params.data_source_name
        self.comparison_results = base_dq_output_helper_params.comparison_results
        self.data_column_collection = base_dq_output_helper_params.data_column_collection
        self.run_type = base_dq_output_helper_params.run_type

    @abc.abstractmethod
    def execute(self) -> BaseDqOutputReport:
        """Execute reporting for one df"""
        raise NotImplementedError

    @staticmethod
    def _get_signal_details(dq_comparison_sample_report_input: DqComparisonSampleReportInput) \
            -> List[SignalDetails]:
        """Return alerts"""

        signals: List[SignalDetails] = []
        if dq_comparison_sample_report_input is not None and \
                dq_comparison_sample_report_input.dq_comparison_result is not None:
            for signal in dq_comparison_sample_report_input.dq_comparison_result:
                signal_details = SignalDetails()
                signal_details.data_entity_name = signal.data_entity_name
                if signal.data_entity == DataEntityType.TABLE:
                    signal_details.column_type = DataEntityType.TABLE.name
                else:
                    col = dq_comparison_sample_report_input.dq_data_column_collection.get_column(
                        signal.data_entity_name)
                    if col is not None:
                        signal_details.column_type = col.column_native_type.name
                signal_details.signal = signal.result_code.name
                signals.append(signal_details)
        return signals

    def _get_result_with_type(self):
        """Get result with type : table comparison result, numeric comparison result, nominal comparison result"""
        table_comparison_result = []
        numeric_comparison_result = []
        nominal_comparison_result = []
        for comparison_result in self.comparison_results:
            if comparison_result.data_entity == DataEntityType.TABLE:
                table_comparison_result.append(comparison_result)
            else:
                data_column = self.data_column_collection.get_column(comparison_result.data_entity_name)
                if data_column.column_native_type == ColumnNativeType.Numeric:
                    numeric_comparison_result.append(comparison_result)
                elif data_column.column_native_type in DqConstants.NOMINAL_COLUMN_TYPES:
                    nominal_comparison_result.append(comparison_result)

        return table_comparison_result, numeric_comparison_result, nominal_comparison_result

    def _get_result_dict(self, comparison_result_list: List[DqComparisonResult]) -> \
            Dict[str, List[DqComparisonResultCode]]:
        """Return column- result code result"""
        result: Dict[str, List[DqComparisonResultCode]] = {}
        for comparison_result in comparison_result_list:
            column_name = self.data_column_collection \
                .get_column(comparison_result.data_entity_name).column_name
            if column_name in result:
                result[column_name].append(comparison_result.result_code)
            else:
                result[column_name] = [comparison_result.result_code]
        return result

    @classmethod
    def _get_full_filter_string(cls, partition: List[PartitionInfo], filter_str: str):
        partition_str = cls._get_partition_string(partition)
        if string_helper.is_null_or_empty(partition_str) and string_helper.is_null_or_empty(filter_str):
            return ""
        if not string_helper.is_null_or_empty(partition_str):
            if not string_helper.is_null_or_empty(filter_str):
                return f"Partition:{partition_str} - Filtered By:'{filter_str}'"
            return partition_str
        return f"Filtered By:'{filter_str}'"

    @classmethod
    def _get_partition_string(cls, partition: List[PartitionInfo]) -> str:
        if list_helper.is_null_or_empty(partition):
            return ""
        return " & ".join([partition_info.to_str() for partition_info in partition])



"""
This module includes AfeApplicationOperations class.
"""
import threading

from organon.afe.core.businessobjects.afe_static_objects import AfeStaticObjects
from organon.fl.generic.interaction.fl_initializer import FlInitializer


class AfeApplicationOperations:
    """
    AfeApplicationOperations
    """

    APPLICATION_INITIALIZED = False
    __INITIALIZATION_LOCK = threading.Lock()

    @classmethod
    def initialize_app(cls):
        """Initializes application."""
        with AfeApplicationOperations.__INITIALIZATION_LOCK:
            if not AfeApplicationOperations.APPLICATION_INITIALIZED:
                cls._on_init()
                AfeApplicationOperations.APPLICATION_INITIALIZED = True

    @classmethod
    def _on_init(cls):
        cls._initialize_fl()
        cls.register_types()
        AfeStaticObjects.set_defaults()

    @classmethod
    def _initialize_fl(cls):
        FlInitializer.application_initialize()

    @classmethod
    def finalize_app(cls):
        """Finalizes application."""

    @classmethod
    def register_types(cls):
        """
        Registers ioc items
        :return: nothing
        """



"""
TO CREATE A WHEEL FILE RUN:
-pip install wheel setuptools
-python setup.py bdist_wheel
"""

from Cython.Build import cythonize
from setuptools import find_packages, setup

from cython_setup import ext_modules

if __name__ == "__main__":
    with open("README.md", "r") as fh:
        long_description = fh.read()

    setup(
        name="autonon",
        version="0.4.0",
        author="Organon Analytics",
        author_email="support@organonanalytics.com",
        description="Organon Automated ML Platform",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://gitlab.com/organon-os/autonon",
        packages=find_packages(exclude=["organon.tests*"]),
        package_data={
            "organon.afe": [
                "data/*"
            ]
        },
        install_requires=[
            "numpy>=1.20.0",
            "pandas>=1.1.1",
            "paramiko>=2.7.2",
            "psutil>=5.7.2",
            "sortedcontainers>=2.3.0",
            "pycryptodome>=3.9.9",
            "plotly>=4.14.1",
            "matplotlib>=3.3.3",
            "scipy>=1.6.0",
            "cryptography>=3.3.1",
            "joblib>=1.0.1",
            "scikit-learn>=1.1",
            "lightgbm>=3.2.1",
            "python-dateutil>=2.8.2",
            "six>=1.16.0",
            "seaborn>=0.11.2",
            "xgboost>=1.6.1",
            "scikit-optimize>=0.9.0",
            "statsmodels >= 0.13.2",
            "tensorflow>=2.10",
            "defusedxml >= 0.6.0",
            "transformers>=4.26.0",
            "datasets>=2.6.1",
            "keras>=2.10.0",
            "ultralytics>=8.0.59"
        ],
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        python_requires='>=3.8',
        ext_modules=cythonize(ext_modules)
    )



"""Includes DqFileRecordSource class."""
from organon.idq.domain.businessobjects.record_sources.dq_base_record_source import DqBaseRecordSource
from organon.idq.domain.enums.dq_record_source_type import DqRecordSourceType


class DqFileRecordSource(DqBaseRecordSource[str]):
    """Dq record source for text file objects."""

    def __init__(self, locator: str):
        super().__init__(locator)

    def get_name(self) -> str:
        return f"File({self.locator})"

    @staticmethod
    def get_type() -> DqRecordSourceType:
        return DqRecordSourceType.TEXT



"""Includes UserInputValidationService class."""
from typing import List, Dict, Generic, TypeVar, Union

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.core.helpers import list_helper
from organon.fl.core.helpers.string_helper import is_null_or_empty
from organon.fl.logging.helpers.log_helper import LogHelper
from organon.idq.core.dq_constants import DqConstants
from organon.idq.services.user_settings.base_dq_user_input import BaseDqUserInput
from organon.idq.services.user_settings.user_calculation_params import UserCalculationParams
from organon.idq.services.user_settings.user_input_source_settings import UserInputSourceSettings
from organon.idq.services.user_settings.user_multi_partition_calculation_params import \
    UserMultiPartitionCalcParams

DqUserInputType = TypeVar("DqUserInputType", bound=BaseDqUserInput)


class BaseUserInputValidationService(Generic[DqUserInputType]):
    """Service for validating settings in DqUserInput"""

    def __init__(self, user_input: DqUserInputType):
        self.user_input: DqUserInputType = user_input

    def validate(self):
        """Executes validation and returns columns and their native types in test input source"""
        self._validate_common_comparison_params()

    def _validate_common_comparison_params(self):
        self._check_range(self.user_input.minimum_cardinality, 0, 100, "Minimum cardinality")
        self._check_range(self.user_input.traffic_light_threshold_green, 0, 0.2, "Traffic Light Threshold Green")
        self._check_range(self.user_input.traffic_light_threshold_yellow, 0.2, 0.4, "Traffic Light Threshold Yellow")
        self._check_range(self.user_input.psi_threshold_green, 0, 0.1, "Psi Threshold Green")
        self._check_range(self.user_input.psi_threshold_yellow, 0.1, 0.2, "Psi Threshold Yellow")
        self._check_range(self.user_input.z_score, 0, 5, "Z Score")
        self._validate_test_groups_info()

    def _validate_included_excluded_columns(self, source_columns: List[List[str]]):
        included_cols = self.user_input.included_column_names
        excluded_cols = self.user_input.excluded_column_names
        included_cols = included_cols if included_cols is not None else []
        excluded_cols = excluded_cols if excluded_cols is not None else []
        common_cols = [col for col in included_cols if col in excluded_cols]
        if len(common_cols) > 0:
            raise KnownException(f"Columns {common_cols} are given in both "
                                 f"included_column_names and excluded_columns_names")
        for included_col in included_cols:
            for columns in source_columns:
                if included_col not in columns:
                    raise KnownException(f"Column '{included_col}' (in included_column_names) does not exist "
                                         f"in all data sources")

    def _validate_comparison_column_info(self, source_column_names: List[str]):
        if self.user_input.column_benchmark_horizons is not None:
            non_existing_comp_col_names = [col for col in self.user_input.column_benchmark_horizons
                                           if col not in source_column_names]
            if len(non_existing_comp_col_names) > 0:
                raise KnownException(f"Following comparison columns(given in column_benchmark_horizons) do not exist"
                                     f" in test calculation input source: {non_existing_comp_col_names}")
            num_calcs = self._get_number_of_calculations()
            for col, horizon in self.user_input.column_benchmark_horizons.items():
                if horizon > num_calcs - 1:
                    LogHelper.warning(f"There are {num_calcs - 1} control calculations but benchmark horizon is "
                                      f"given {horizon} for column '{col}'")

        if self.user_input.duplicate_control_columns is not None:
            non_existing_comp_col_names = [col for col in self.user_input.duplicate_control_columns
                                           if col not in source_column_names]
            if len(non_existing_comp_col_names) > 0:
                raise KnownException(f"Following comparison columns(given in duplicate_control_columns) do not exist "
                                     f"in test calculation input source: {non_existing_comp_col_names}")

    def _get_eligible_column_types(self):
        col_types = [ColumnNativeType.Numeric] + DqConstants.NOMINAL_COLUMN_TYPES
        if self.user_input.include_date_columns is False and ColumnNativeType.Date in col_types:
            col_types.remove(ColumnNativeType.Date)
        return col_types

    def _validate_test_groups_info(self):
        if self.user_input.test_group_benchmark_horizons is not None:
            num_calcs = self._get_number_of_calculations()
            for test_group, horizon in self.user_input.test_group_benchmark_horizons.items():
                if horizon > num_calcs - 1:
                    LogHelper.warning(f"There are {num_calcs - 1} control calculations but benchmark horizon is "
                                      f"given {horizon} for test group '{test_group}'")

    def _get_number_of_calculations(self) -> int:
        calc_params = self.user_input.calc_params
        if list_helper.is_null_or_empty(calc_params.partitions):
            return 1
        return len(calc_params.partitions)

    def _validate_column_metadata_list(self, source_column_names: List[str]):
        if self.user_input.column_default_values is not None:
            non_existing_comp_col_names = [col for col in self.user_input.column_default_values
                                           if col not in source_column_names]
            if len(non_existing_comp_col_names) > 0:
                raise KnownException(f"Following columns(given in column_default_values) do not exist in "
                                     f"test calculation input source: {non_existing_comp_col_names}")

        if self.user_input.excluded_column_names is not None:
            non_existing_comp_col_names = [col for col in self.user_input.excluded_column_names
                                           if col not in source_column_names]
            if len(non_existing_comp_col_names) > 0:
                raise KnownException(f"Following columns(given in excluded_column_names) do not exist in "
                                     f"test calculation input source: {non_existing_comp_col_names}")

    def _validate_calculation_parameters(self, _index: int,
                                         calculation_parameters: Union[UserMultiPartitionCalcParams,
                                                                       UserCalculationParams],
                                         column_native_types: Dict[str, ColumnNativeType]):
        self.validate_calculation_input_source_settings(calculation_parameters.input_source_settings)
        allowed_native_types = self._get_eligible_column_types()
        eligible_column_list = [col for col, native_type in column_native_types.items()
                                if native_type in allowed_native_types]
        if list_helper.is_null_or_empty(eligible_column_list):
            if _index is not None:
                raise KnownException(f"No eligible columns found in data source of Calculation-{_index + 1}")
            raise KnownException("No eligible columns found in data source")

        self._check_range(calculation_parameters.max_nominal_cardinality_count, 0, 100000,
                          "Max. nominal cardinality count")

        if calculation_parameters.outlier_parameter is not None and \
                calculation_parameters.outlier_parameter not in [1.5, 3.0]:
            raise KnownException("Outlier parameter must be either 1.5 or 3")

    @classmethod
    def validate_calculation_input_source_settings(cls, settings: UserInputSourceSettings):
        """Validates input source settings in calculation parameters"""
        if settings.source is None:
            raise KnownException("Input source cannot be None")
        if isinstance(settings.source, str):
            if is_null_or_empty(settings.source.strip()):
                raise KnownException("Input file name is empty")

        cls._check_range(settings.sampling_ratio, 0, 1.0, "Sampling ratio")

    @classmethod
    def _check_range(cls, value: float, min_val: float, max_val: float, definition: str):
        if value is not None and (value < min_val or value > max_val):
            raise KnownException(f"{definition} should be between {min_val} and {max_val}")



"""Includes UserCalculationParams class."""
from typing import List

from organon.idq.services.user_settings.base_user_calculation_params import BaseUserCalculationParams
from organon.idq.services.user_settings.user_partition_info import UserPartitionInfo


class UserCalculationParams(BaseUserCalculationParams):
    """Calculation settings user class"""

    def __init__(self, calc_name, input_source_settings, max_nom_card_count, outlier_param, use_pop_nom_stats,
                 partition=None):
        super().__init__(calc_name, input_source_settings, max_nom_card_count, outlier_param, use_pop_nom_stats)
        self.partition: List[UserPartitionInfo] = partition



"""Includes MultiTargetModellingService class."""
from typing import List, Optional

from organon.afe.domain.modelling.base_modelling_service import BaseModellingService
from organon.afe.domain.modelling.businessobjects.data_frame_builder import DataFrameBuilder
from organon.afe.domain.modelling.supervised.afe_supervised_algorithm_settings import AfeSupervisedAlgorithmSettings
from organon.afe.domain.modelling.supervised.multi_target_entity_container_service import \
    MultiTargetEntityContainerService
from organon.afe.domain.modelling.supervised.multi_target_modelling_helper import MultiTargetModellingHelper
from organon.afe.domain.settings.afe_modelling_settings import AfeModellingSettings
from organon.fl.core.businessobjects.dict_dataframe import DictDataFrame
from organon.fl.core.businessobjects.idata_partition import IDataPartition


class MultiTargetModellingService(BaseModellingService):
    """Class for AFE modelling with multiple targets."""

    def __init__(self, settings: AfeModellingSettings):
        super().__init__(settings)
        self.modelling_helper = MultiTargetModellingHelper(self._settings)

    @staticmethod
    def _get_partition(frame_builder: DataFrameBuilder = None,
                       algorithm_settings: AfeSupervisedAlgorithmSettings = None) -> Optional[IDataPartition]:
        partition = MultiTargetModellingHelper.get_partition(frame_builder, algorithm_settings)
        return partition

    def _get_selected_cols(self, num_threads: int,
                           partition: IDataPartition = None,
                           algorithm_settings: AfeSupervisedAlgorithmSettings = None,
                           frame_builder: DataFrameBuilder = None,
                           frame_with_all_columns: DictDataFrame = None,
                           is_final=False) -> List[str]:
        return self.modelling_helper.get_selected_cols(num_threads, partition, algorithm_settings, frame_builder,
                                                       frame_with_all_columns, is_final)

    def _get_final_columns_metrics(self, num_threads: int, frame_builder: DataFrameBuilder,
                                   frame_with_all_columns: DictDataFrame, reduced_columns: List[str],
                                   algorithm_settings: AfeSupervisedAlgorithmSettings) -> dict:
        return self.modelling_helper.get_final_columns_metrics(num_threads, frame_builder, frame_with_all_columns,
                                                               reduced_columns, algorithm_settings)

    def control_target_record_files(self, multi_target_entity_container_service: MultiTargetEntityContainerService):
        self.modelling_helper.control_target_record_files(multi_target_entity_container_service)



"""Includes CrossValidationSettings class"""
from dataclasses import dataclass
from typing import List, Optional

import pandas as pd

from organon.ml.modelling.algorithms.core.abstractions.base_modeller import BaseModeller


@dataclass
class CrossValidationSettings:
    """Settings for CrossValidation"""
    modellers: List[BaseModeller]
    train_data: pd.DataFrame
    target_data: pd.Series
    cv_count: int
    bin_count: Optional[int]
    return_test_fold: bool



"""Includes BaseSupervisedFeatureSelectionService class."""
import abc
from typing import TypeVar, Generic

import pandas as pd
from sklearn.preprocessing import LabelEncoder

from organon.ml.common.enums.target_type import TargetType
from organon.ml.feature_selection.domain.objects.base_selection_output import BaseSelectionOutput
from organon.ml.feature_selection.domain.objects.settings.base_supervised_feature_selection_settings import \
    BaseSupervisedFeatureSelectionSettings
from organon.ml.feature_selection.domain.services.base_feature_selection_service import BaseFeatureSelectionService

SupervisedFeatureSelectionSettingsType = TypeVar("SupervisedFeatureSelectionSettingsType",
                                                 bound=BaseSupervisedFeatureSelectionSettings)
FeatureSelectionOutputType = TypeVar("FeatureSelectionOutputType", bound=BaseSelectionOutput)


class BaseSupervisedFeatureSelectionService(Generic[SupervisedFeatureSelectionSettingsType, FeatureSelectionOutputType],
                                            BaseFeatureSelectionService[SupervisedFeatureSelectionSettingsType,
                                                                        FeatureSelectionOutputType],
                                            metaclass=abc.ABCMeta):
    """Base service class for supervised feature selection services"""

    @classmethod
    def _get_target_data_as_series_of_numbers(cls, settings: SupervisedFeatureSelectionSettingsType):
        if settings.target_type == TargetType.BINARY:
            label_encoder = LabelEncoder()
            label_encoder.fit(settings.target)
            return pd.Series(label_encoder.transform(settings.target[settings.target.columns[0]]))
        return settings.target[settings.target.columns[0]]



"""
This module includes AfeUnsupervisedAlgorithmSettings class.
"""
from organon.afe.domain.common.reader_helper import get_values_from_kwargs
from organon.afe.domain.settings.afe_algorithm_settings import AfeAlgorithmSettings


class AfeUnsupervisedAlgorithmSettings(AfeAlgorithmSettings):
    """
    Algorithm settings for Automated Feature Extraction with Unsupervised Learning type
    """
    ATTR_DICT = {
        "bin_count": int,
        "r_factor": float,
        "random_state": int,
        "max_column_count": int,
        "is_logging": bool
    }
    ATTR_DICT.update(AfeAlgorithmSettings.ATTR_DICT)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.bin_count: int = None
        self.r_factor: float = None
        self.random_state: int = None
        self.max_column_count: int = None
        self.is_logging: bool = None

        get_values_from_kwargs(self, self.ATTR_DICT, kwargs)



"""Includes SelectionMetrics enum class"""
from enum import Enum, auto


class SelectionMetrics(Enum):
    """Best Model Selection Metrics for Grid Search"""
    VAL_LOSS = auto()
    ROC_AUC = auto()
    ACC_SCORE = auto()
    F1_SCORE = auto()



""" todo docstring """
import math
from typing import List

import numpy as np
from sklearn import metrics
from sklearn.ensemble import RandomForestRegressor

from organon.idq.domain.controls.helpers.algorithms.base_predictor import BasePredictor
from organon.idq.domain.controls.helpers.algorithms.prediction_result import PredictionResult


class RandomForestPredictor(BasePredictor):
    """ todo docstring """

    @staticmethod
    def predict(input_data: List[float], t_val: float, n_estimators=200, max_depth=10,
                max_samples_leaf=1) -> PredictionResult:
        """randomforestregressor prediction, input_data from t-n to t-1"""
        rf_ = RandomForestRegressor(n_estimators=n_estimators, max_depth=max_depth, min_samples_leaf=max_samples_leaf)
        features, labels = RandomForestPredictor.generate_train_data_with_lags(input_data, t_val)
        impute_value = float(np.finfo(np.float32).min)
        normalized_features = np.nan_to_num(features, nan=impute_value)
        features_to_fit = features[:-1]
        labels_to_fit = labels[:-1]
        nan_indices = [i for i, val in enumerate(labels_to_fit) if math.isnan(val)]
        features_nan_target_removed = np.nan_to_num(np.delete(features_to_fit, nan_indices, axis=0), nan=impute_value)
        labels_nan_removed = [val for val in labels_to_fit if not math.isnan(val)]
        rf_.fit(features_nan_target_removed, labels_nan_removed)

        predictions = rf_.predict(normalized_features)
        pred_result = PredictionResult(labels)
        pred_result.predictions = predictions
        past_values_predictions = np.delete(predictions[:-1], nan_indices, axis=0)
        pred_result.mean_square_err = metrics.mean_squared_error(labels_nan_removed, past_values_predictions)
        return pred_result



""" This module includes ComparisonHelperFunctions"""
import math
from typing import Tuple

from organon.idq.domain.enums.signal_type import SignalType


class ComparisonHelperFunctions:
    """ Class for ComparisonHelperFuncstions """

    @staticmethod
    def get_traffic_light_signal(change: float, yellow_threshold: float, green_threshold: float) -> SignalType:
        """ :returns signal type"""
        if change is None:
            return SignalType.GREEN
        if change < green_threshold:
            return SignalType.GREEN
        if change < yellow_threshold:
            return SignalType.YELLOW
        return SignalType.RED

    @staticmethod
    def get_percentage_change(current: float, reference: float) -> float:
        """:returns percentage change of current value"""
        delta = reference - current
        if current == 0 and reference == 0:
            return 0
        if reference == 0:
            return 1
        return abs(delta / reference)

    @staticmethod
    def compute_control_mean_std_avg(data: list) -> Tuple[float, float]:
        """computes mean and std for past data"""
        weight = 0.0
        variance = 0.0
        mean = 0.0
        val_sum = 0.0
        for val in data:
            delta = val - mean
            weight += 1.0
            variance += delta * delta * (1.0 - (1.0 * 1.0) / weight)
            mean += delta * (1.0 / weight)
            val_sum += val
        if weight >= 2.0:
            variance = variance / (weight - 1)
        std_error = math.sqrt(variance) / math.sqrt(len(data))
        return mean, std_error

    @staticmethod
    def is_outside_confidence_interval(z_score: float, control_mean: float, control_std_err: float,
                                       test_mean: float) -> bool:
        """ColumnMeanIsOutsideMeanConfidenceInterval control"""
        upper_limit = control_mean + z_score * control_std_err
        lower_limit = control_mean - z_score * control_std_err
        if test_mean < lower_limit or test_mean > upper_limit:
            return True
        return False



"""Includes CFSSelectionSettings class."""
import pandas as pd

from organon.ml.common.enums.target_type import TargetType
from organon.ml.feature_selection.domain.objects.settings.base_supervised_feature_selection_settings import \
    BaseSupervisedFeatureSelectionSettings


class CFSSelectionSettings(BaseSupervisedFeatureSelectionSettings):
    """Settings for CFS selection service"""

    def __init__(self, data: pd.DataFrame, target: pd.DataFrame, target_type: TargetType, max_backtracks=5,
                 prioritize_performance=False):
        super().__init__(data, target, target_type)
        self.max_backtracks = max_backtracks
        self.prioritize_performance = prioritize_performance



"""
This module includes BaseAfeOutputReport class.
"""
from datetime import datetime
from typing import Dict, List, Generic, TypeVar

from organon.afe.domain.modelling.businessobjects.base_afe_feature import BaseAfeFeature
from organon.afe.domain.reporting.transformation import Transformation


class FeatureCountReport:
    """Stores feature count statistics grouped by different fields"""

    def __init__(self, by_dimension: Dict[str, int], by_quantity: Dict[str, int], by_operator: Dict[str, int],
                 by_time_window: Dict[str, int]):
        self.by_dimension = by_dimension
        self.by_quantity = by_quantity
        self.by_operator = by_operator
        self.by_time_window = by_time_window


AfeFeatureType = TypeVar("AfeFeatureType", bound=BaseAfeFeature)


class BaseAfeOutputReport(Generic[AfeFeatureType]):
    """Afe output information to be reported."""

    def __init__(self):
        self.model_identifier: str = None
        self.output_features: Dict[str, AfeFeatureType] = None
        self.all_features: Dict[str, AfeFeatureType] = None
        self.dimension_names_map: Dict[str, Dict[int, str]] = None
        self.transformations: Dict[str, Transformation] = None
        self.execution_time: float = None
        self.memory_usage: Dict[datetime, float] = None
        self.cpu_usage: Dict[datetime, float] = None
        self.event_times: Dict[str, datetime] = None
        self.feature_counts_report: FeatureCountReport = None
        self.final_column_metrics: List[dict] = None



"""Includes BaseSelectionOutput class."""
from typing import List


class BaseSelectionOutput:
    """Base output for feature selection services"""
    def __init__(self):
        self.selected_features: List[str] = None



"""Includes XGBoostRegressor class."""
from xgboost import XGBRegressor

from organon.ml.modelling.algorithms.core.abstractions.base_sklearn_regressor import BaseSklearnRegressor


class XGBoostRegressor(BaseSklearnRegressor[XGBRegressor]):
    """Wrapper for XGBRegressor"""

    @property
    def regressor_class(self):
        return XGBRegressor



"""This module includes helper functions for retrieving function definition info."""
import inspect
from typing import Tuple, List, Dict, Any


def get_parameters(func) -> Tuple[List[str], Dict[str, Any]]:
    """
    Returns parameters of given func
    :param func:
    :return: tuple --> [parameter name list , parameters with default values as dictionary]
    """
    signature = inspect.signature(func)
    params_with_defaults = {key: value.default for key, value in signature.parameters.items()
                            if value.default != inspect.Parameter.empty}
    return list(signature.parameters.keys()), params_with_defaults



"""
This module includes TargetFileRecordCollection class.
"""
from random import Random
from typing import List

import numpy as np

from organon.afe.domain.modelling.businessobjects.target_file_record import TargetFileRecord


class TargetFileRecordCollection:
    """
    Collection of target file records
    """

    def __init__(self, max_record_count: int):
        super().__init__()
        self.max_record_count: int = max_record_count
        self.__sampled: bool = None
        self.__last_filled_index: int = None
        self.__indices: List[int] = None
        self.__entities: np.array = np.full(max_record_count, None, dtype=np.chararray)
        self.__event_dates: np.array = np.full(max_record_count, -1, dtype=np.int64)
        self.__target_binary_values: np.array = np.full(max_record_count, -1, dtype=np.int8)
        self.__target_scalar_values: np.array = np.full(max_record_count, None, dtype=np.float64)
        self.__target_multi_class_values: np.array = np.full(max_record_count, None, dtype=np.object_)

    @property
    def actual_record_count(self) -> int:
        """Number of rows that was filled with data.(Ignoring rows with dummy data)"""
        if self.__last_filled_index is None:
            return 0
        return self.__last_filled_index + 1

    @property
    def sampled(self) -> bool:
        """Returns value of private attribute '__sampled'"""
        return self.__sampled

    @property
    def indices(self) -> List[int]:
        """Returns value of private attribute '__indices'"""
        return self.__indices

    @property
    def entities(self) -> np.ndarray:
        """Returns value of private attribute '__entities'"""
        return self.__entities

    @property
    def event_dates(self) -> np.ndarray:
        """Returns value of private attribute '__event_dates'"""
        return self.__event_dates

    @property
    def target_binary_values(self) -> np.ndarray:
        """Returns value of private attribute '__target_binary_values'"""
        return self.__target_binary_values

    @property
    def target_scalar_values(self) -> np.ndarray:
        """Returns value of private attribute '__target_scalar_values'"""
        return self.__target_scalar_values

    @property
    def target_multi_class_values(self) -> np.ndarray:
        """Returns value of private attribute '__target_scalar_values'"""
        return self.__target_multi_class_values

    @property
    def sampled_count(self) -> int:
        """Returns sampled_count"""
        return len(self.__indices) if self.sampled else self.actual_record_count

    def __get_unique_entities(self) -> List[str]:
        """
        Returns list of unique entity IDs in collection
        :return: List[str]
        """
        _list = []
        for i in range(self.actual_record_count):
            _list.append(self.__entities[i])
        return list(set(_list))

    def get_sampled_unique_entities(self, random_generator: Random, max_entities: int) -> List[str]:
        """
        Returns a list of sampled unique entity IDs
        :param random_generator: Random number generator
        :type random_generator: Random
        :param max_entities:
        :type max_entities: int
        :return: List[str]
        """
        if self.actual_record_count <= max_entities:
            self.__sampled = False
            self.__indices = None
            return self.__get_unique_entities()

        self.__sampled = True
        _list = []
        for i in range(self.actual_record_count):
            _list.append((random_generator.random(), i))
        _list.sort()
        self.__indices = []
        samples = []
        for i in range(max_entities):
            pair = _list[i]
            index = pair[1]
            self.__indices.append(index)
            samples.append(self.entities[index])
        return list(set(samples))

    def append(self, record: TargetFileRecord):
        """Adds given record to record collection"""
        if self.__last_filled_index is None:
            self.__last_filled_index = -1
        self.__last_filled_index += 1
        index = self.__last_filled_index
        self.__entities[index] = record.entity_id
        self.__event_dates[index] = record.event_date
        self.__target_binary_values[index] = record.target_binary.value
        self.__target_scalar_values[index] = record.target_scalar
        self.__target_multi_class_values[index] = record.target_multi_class



"""
This module keeps Histogram16 class.
"""
from typing import TypeVar, Dict, List, Generic
import numpy as np
from organon.fl.core.helpers.dict_helper import add_if_key_not_exist
from organon.fl.core.collections.pair import Pair
T1 = TypeVar('T1')


class Histogram16(Generic[T1]):
    """
    This class keeps the features of histogram and the functions used in histograms such as build indices.
    """
    def __init__(self):
        self.__frequencies: Dict[T1, np.int64] = {}
        self.__index: Dict[T1, np.int16] = {}
        self.__reverse_index: Dict[np.int16, T1] = {}
        self.__compressed_index: Dict[T1, np.int16] = {}
        self.__compressed_reverse_index: Dict[np.int16, List[T1]] = {}
        self.__original_2_compressed: Dict[np.int16, np.int16] = {}
        self.__bulk_set_index: np.int16 = np.int16(-1)
        self.__bulk_set: List[T1] = []

    @property
    def frequencies(self) -> Dict[T1, np.int64]:
        """
        Returns the frequencies dictionary.
        :return: frequencies dictionary
        """
        return self.__frequencies

    @property
    def index(self) -> Dict[T1, np.int16]:
        """
        Returns the index dictionary.
        :return: index dictionary
        """
        return self.__index

    @property
    def reverse_index(self) -> Dict[np.int16, T1]:
        """
        Returns the reverse_index dictionary.
        :return: reverse_index dictionary
        """
        return self.__reverse_index

    @property
    def compressed_index(self) -> Dict[T1, np.int16]:
        """
        Returns the compressed_index dictionary.
        :return: compressed_index dictionary
        """
        return self.__compressed_index

    @property
    def compressed_reverse_index(self) -> Dict[np.int16, List[T1]]:
        """
        Returns the compressed_reverse_index dictionary.
        :return: compressed_reverse_index dictionary
        """
        return self.__compressed_reverse_index

    @property
    def original_to_compressed(self) -> Dict[np.int16, np.int16]:
        """
        Returns the original_2_compressed dictionary.
        :return: original_2_compressed dictionary
        """
        return self.__original_2_compressed

    @property
    def bulk_set_index(self) -> np.int16:
        """
        Returns bulk_set_index value.
        :return: bulk_set_index value
        """
        return self.__bulk_set_index

    @property
    def bulk_set(self) -> List[T1]:
        """
        Returns bulk_set_index.
        :return: bulk_set_index
        """
        return self.__bulk_set

    def get_index_value(self, key: T1) -> np.int16:
        """
        Returns the key's value in index dictionary.
        :param key: key of index dictionary
        :return: the key's value in index dictionary
        """
        return self.__index[key]

    def build_indices(self):
        """
        Builds the indices of the histogram
        :return: nothing
        """
        freq_length: int = len(self.__frequencies)
        list1: List[Pair[np.float64, T1]] = [None] * freq_length
        frequency_sum: np.float64 = np.float64(0.0)
        index: int = 0
        for item, value in self.__frequencies.items():
            list1[index] = (Pair(value, item))
            frequency_sum += self.__frequencies[item]
            index += 1
        list1.sort()

        list2: List[Pair[np.float64, T1]] = [None] * freq_length
        index = 0
        for pair in reversed(list1):
            list2[index] = Pair(pair.first / frequency_sum, pair.second)
            index += 1
        for pair in list2:
            percentage: np.float64 = pair.first
            key: T1 = pair.second
            if percentage >= 0.01:
                add_if_key_not_exist(key, self.__index[key], self.__compressed_index)
                add_if_key_not_exist(self.__index[key], [key], self.__compressed_reverse_index)
            else:
                if self.__bulk_set_index not in self.__compressed_reverse_index:
                    self.__compressed_reverse_index[self.__bulk_set_index] = []
                add_if_key_not_exist(key, self.__bulk_set_index, self.__compressed_index)
                self.__compressed_reverse_index[self.__bulk_set_index].append(key)
                self.__bulk_set.append(key)

        keys: List[T1] = list(self.__index.keys())
        for key in keys:
            add_if_key_not_exist(self.__index[key], self.__compressed_index[key],
                                 self.__original_2_compressed)

    def build_indices_with_params(self, compression_ration: float, max_cardinality: int):
        """
        Builds the indices of the histogram according to the compression ratio and max cardinality
        :param compression_ration: measurement of the relative reduction in size of data representation
        produced by a data compression algorithm
        :param max_cardinality: maximum number of entity instances that can participate in a relationship
        :return: nothing
        """
        freq_length: int = len(self.__frequencies)
        list1: List[Pair[np.float64, T1]] = [None] * freq_length
        index: int = 0
        for item, value in self.__frequencies.items():
            list1[index] = Pair(value, item)
            # list1.append(Pair(self.__frequencies[item], item))
            index += 1

        list1.sort()
        list2: List[Pair[np.float64, T1]] = [None] * freq_length
        last_pair: Pair[np.float64, T1] = list1[freq_length-1]
        list2[0] = Pair(last_pair.first, last_pair.second)
        grand_sum: np.float64 = last_pair.first
        if len(list1) > 1:
            index = 1
            reversed_list1 = reversed(list1[:-1])
            for pair in reversed_list1:
                grand_sum += pair.first
                list2[index] = Pair(grand_sum, pair.second)
                index += 1

        # list1.clear()
        list3: List[Pair[np.float64, T1]] = [None] * freq_length
        index = 0
        for pair in list2:
            list3[index] = Pair(pair.first / grand_sum, pair.second)
            index += 1

        # list2.clear()
        found: bool = False
        index: int = 0
        for pair in list3:
            percentage: np.float64 = pair.first
            key: T1 = pair.second
            # mark: np.int16 = np.int16(list3.index(pair))
            # self.index.update({key: mark})
            if percentage <= compression_ration and index < max_cardinality:
                add_if_key_not_exist(key, self.__index[key], self.__compressed_index)
                add_if_key_not_exist(self.__index[key], [key], self.__compressed_reverse_index)
            else:
                if not found:
                    found = True
                    # self.bulk_set_index = mark
                    add_if_key_not_exist(key, self.__index[key], self.__compressed_index)
                    add_if_key_not_exist(self.__index[key], [key], self.__compressed_reverse_index)
                else:
                    if self.__bulk_set_index not in self.__compressed_reverse_index:
                        self.__compressed_reverse_index[self.__bulk_set_index] = []
                    add_if_key_not_exist(key, self.__bulk_set_index, self.__compressed_index)
                    self.__compressed_reverse_index[self.__bulk_set_index].append(key)
                    self.__bulk_set.append(key)
            index += 1
        keys: List[T1] = list(self.__index.keys())
        for key in keys:
            add_if_key_not_exist(self.__index[key], self.__compressed_index[key],
                                 self.__original_2_compressed)

    def add(self, key: T1, frequency: int = None):
        """
        Adds or updates the key's value in the dictionaries.
        :param key: key for index dict, value for reverse index dict
        :param frequency: frequency value
        :return: nothing
        """
        if frequency is None:
            if key in self.__frequencies:
                self.__frequencies[key] += 1
            else:
                self.__frequencies[key] = 1
                index_val: np.int16 = np.int16(len(self.__index))
                add_if_key_not_exist(key, index_val, self.__index)
                add_if_key_not_exist(index_val, key, self.__reverse_index)
        else:
            if key in self.__frequencies:
                self.__frequencies[key] += frequency
            else:
                self.__frequencies[key] = frequency
                index_val: np.int16 = np.int16(len(self.__index))
                add_if_key_not_exist(key, index_val, self.__index)
                add_if_key_not_exist(index_val, key, self.__reverse_index)



"""Includes SignalDetails class"""
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode


class SignalDetails:
    """Class for signal details """

    def __init__(self):
        self.data_entity_name: str = None
        self.column_type: str = None
        self.signal: DqComparisonResultCode = None

    def to_dict(self):
        """return dict for signal detail"""
        return {
            'data_entity_name': self.data_entity_name,
            'column_type': self.column_type,
            'signal': self.signal
        }



"""
This module includes TargetFileRecord class.
"""
import numpy as np

from organon.afe.domain.enums.binary_target_class import BinaryTargetClass


class TargetFileRecord:
    """
    Class corresponding to a row in target file
    """

    def __init__(self):
        self.entity_id: str = None
        self.event_date: int = None
        self.target_scalar: np.float64 = None
        self.target_binary: BinaryTargetClass = None
        self.target_multi_class: np.object_ = None



"""Includes UserPartitionInfo class."""
from typing import Union, List

from organon.idq.services.user_settings.user_date_value_definition import UserDateValueDefinition


class UserPartitionInfo:
    """User settings class for PartitionInfo."""

    def __init__(self):
        self.column_name: str = None
        self.column_values: Union[List[UserDateValueDefinition], List[float], List[str]] = None



"""Includes helper functions used for multiprocessing/multithreading."""
from threading import Thread
from typing import List, Callable, Any, Optional

from joblib import Parallel, delayed


def execute_parallel(_list: Optional[List[Any]], func: Callable, num_jobs=-1, backend: str = None,
                     require_shared_memory: bool = False, prefer_threads: bool = None,
                     **kwargs):
    """
    Runs given 'func' multiple times using given number of threads/processes, with elements in '_list' as arguments.

    :param _list: Includes arguments for every execution of 'func'. That is, func will be run len(_list) times.
    :param func: Function to execute in parallel.
    :param num_jobs: Number of threads/processes to use
    :param backend: 'joblib' library backend type to use.Values: loky(default), threading, multiprocessing
    :param require_shared_memory: Should be true if methods use a shared variable.
    :param prefer_threads: If true, it hints that func can efficiently use threads.
    :param kwargs:
    :return: Results of 'func' in the order according to arguments in _list.
    """
    require = "sharedmem" if require_shared_memory else None
    prefer = None
    if prefer_threads:
        prefer = "threads"
    args_list = []
    for arg in _list:
        if not isinstance(arg, tuple):
            arg = (arg,)
        args_list.append(arg)
    return Parallel(n_jobs=num_jobs, backend=backend, require=require, prefer=prefer, **kwargs)(
        delayed(func)(*args) for args in args_list)


def execute_multithreaded(_list: List[Any], func: Callable, num_threads=None):
    """
    Runs given 'func' multiple times using given number of threads, with elements in '_list' as arguments

    :param _list: Includes arguments for every executiın of 'func'. That is, func will be run len(_list) times.
    :param func: Function to execute in parallel.
    :param num_threads: Number of threads to use to execute 'func' in parallel.
    """
    list_len = len(_list)
    if num_threads is None:
        num_threads = list_len
    num_threads = min(num_threads, list_len)
    all_threads = []
    index = 0
    step = int(list_len / num_threads)
    remainder = list_len % num_threads
    for i in range(num_threads):
        last_index = index + step
        if remainder != 0 and i < remainder:
            last_index += 1
        arg_lists = _list[index: last_index]
        thread = Thread(target=__exec_multiple, args=(func, arg_lists))
        thread.start()
        all_threads.append(thread)
        index = last_index
    for thr in all_threads:
        thr.join()


def __exec_multiple(func, arg_lists):
    for arg_list in arg_lists:
        func(*arg_list)



"""
This module includes TargetRecordSource class.
"""
import abc
from typing import TypeVar, Generic

from organon.idq.domain.enums.dq_record_source_type import DqRecordSourceType

DqSourceLocatorType = TypeVar("DqSourceLocatorType")


class DqBaseRecordSource(Generic[DqSourceLocatorType], metaclass=abc.ABCMeta):
    """
    Description of a dq record source.
    """

    def __init__(self, locator: DqSourceLocatorType):
        self.locator: DqSourceLocatorType = locator

    @abc.abstractmethod
    def get_name(self) -> str:
        """Returns name of the record source"""
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def get_type() -> DqRecordSourceType:
        """Returns type of the source"""
        raise NotImplementedError



"""Includes IDQ class."""
import abc
from typing import Optional, List, TypeVar, Generic, Dict

import organon
from organon.common.helpers import dev_mode_helper
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.idq.core.dq_constants import DqConstants
from organon.idq.domain.businessobjects.dq_calculation_output import DqCalculationOutput
from organon.idq.domain.reporting.objects.base_dq_output_report import BaseDqOutputReport
from organon.idq.domain.services.base_user_input_service import BaseUserInputService
from organon.idq.domain.services.dq_full_process_executor import DqFullProcessExecutor
from organon.idq.domain.settings.abstractions.dq_full_process_input import DqFullProcessInput
from organon.idq.services.idq_application_operations import IdqApplicationOperations
from organon.idq.services.user_settings.base_dq_user_input import BaseDqUserInput

UserInputServiceType = TypeVar("UserInputServiceType", bound=BaseUserInputService)
DqUserInputType = TypeVar("DqUserInputType", bound=BaseDqUserInput)


class BaseIDQ(Generic[UserInputServiceType, DqUserInputType], metaclass=abc.ABCMeta):
    """User service class for IDQ."""

    def __init__(self):
        self._initialize_idq()
        self.results: Optional[BaseDqOutputReport] = None
        self.calculation_outputs: Optional[List[DqCalculationOutput]] = None
        self._user_settings: DqUserInputType = self._get_user_settings_instance()

    @abc.abstractmethod
    def _get_user_settings_instance(self) -> DqUserInputType:
        """Returns empty DqUserInput instance"""

    @classmethod
    def _initialize_idq(cls):
        IdqApplicationOperations.initialize_app()

    def execute(self) -> BaseDqOutputReport:
        """Executes idq and returns results"""
        helper = self._get_user_input_service()
        full_process_input = helper.convert_to_full_process_input()
        executor = self._get_dq_full_process_executor(full_process_input)
        results, calculation_results = executor.execute()
        self.results = results
        self.calculation_outputs = calculation_results
        return results

    @classmethod
    def init_dev_mode(cls, log_to_console: bool = True, log_file: str = "application.log"):
        """Initializes development mode."""
        dev_mode_helper.init_dev_mode(organon.idq.__name__, log_to_console=log_to_console, log_file=log_file)

    @abc.abstractmethod
    def _get_user_input_service(self) -> UserInputServiceType:
        """Returns UserInputService instance"""

    @classmethod
    def _get_dq_full_process_executor(cls, full_process_input: DqFullProcessInput) -> DqFullProcessExecutor:
        return DqFullProcessExecutor(full_process_input)

    def _check_results(self):
        if self.results is None:
            raise ValueError("You should run execute function first to get the results.")

    def set_settings(self,
                     column_benchmark_horizons: Dict[str, int] = None,
                     duplicate_control_columns: List[str] = None,
                     maximum_nom_cardinality: int = DqConstants.CONTROL_MAXIMUM_NOM_CARDINALITY_DEFAULT,
                     minimum_cardinality: int = DqConstants.MINIMUM_CARDINALITY_DEFAULT,
                     traffic_light_threshold_yellow: float = DqConstants.TRAFFIC_LIGHT_THRESHOLD_YELLOW,
                     traffic_light_threshold_green: float = DqConstants.TRAFFIC_LIGHT_THRESHOLD_GREEN,
                     psi_threshold_yellow: float = DqConstants.PSI_THRESHOLD_YELLOW,
                     psi_threshold_green: float = DqConstants.PSI_THRESHOLD_GREEN,
                     z_score: float = DqConstants.Z_SCORE_DEFAULT,
                     excluded_test_groups: List[str] = None,
                     test_group_benchmark_horizons: Dict[str, int] = None,
                     included_column_names: List[str] = None,
                     excluded_column_names: List[str] = None,
                     column_default_values: Dict[str, List[str]] = None,
                     include_date_columns: bool = None
                     ):
        """
        Adds comparison settings

        :param column_benchmark_horizons:
        :param duplicate_control_columns:
        :param minimum_cardinality:
        :param maximum_nom_cardinality:
        :param traffic_light_threshold_yellow:
        :param traffic_light_threshold_green:
        :param psi_threshold_yellow:
        :param psi_threshold_green:
        :param z_score:
        :param excluded_test_groups:
        :param test_group_benchmark_horizons:
        :param included_column_names:
        :param excluded_column_names:
        :param column_default_values:
        :param include_date_columns:
        """
        # todo parametre açıklamaları eklenecek
        # pylint: disable=too-many-arguments
        self._user_settings.set_comparison_params(column_benchmark_horizons, duplicate_control_columns,
                                                  maximum_nom_cardinality, minimum_cardinality,
                                                  traffic_light_threshold_yellow, traffic_light_threshold_green,
                                                  psi_threshold_yellow, psi_threshold_green,
                                                  z_score, excluded_test_groups, test_group_benchmark_horizons,
                                                  included_column_names, excluded_column_names, column_default_values,
                                                  include_date_columns)

    def _check_calculation_results(self):
        if self.calculation_outputs is None:
            raise KnownException("No calculation results generated yet.")

    @staticmethod
    def set_secret_key(secret_key: str):
        """Sets secret key for encryption tasks used during serialization etc."""
        DqConstants.set_secret_key(secret_key)



# pylint: skip-file
"""Module Definition."""
import abc
import numpy as np


class IBlas(metaclass=abc.ABCMeta):
    """Class Definition."""
    @staticmethod
    @abc.abstractmethod
    def saxpy(n: int, a: np.single, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def daxpy(n: int, a: np.double, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def saxpby(n: int, a: np.single, x: np.ndarray, incx: int, b: np.single, y: np.ndarray, incy: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def daxpby(n: int, a: np.double, x: np.ndarray, incx: int, b, y: np.ndarray, incy: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def scopy(n: int, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def dcopy(n: int, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def scopynan(n: int, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def dcopynan(n: int, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def sdot(n: int, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def ddot(n: int, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def snrm2(n: int, x: np.ndarray, incx: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def dnrm2(n: int, x: np.ndarray, incx: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def sscal(n: int, a: np.single, x: np.ndarray, incx: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def dscal(n: int, a: np.double, x: np.ndarray, incx: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def sswap(n: int, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def dswap(n: int, x: np.ndarray, incx: int, y: np.ndarray, incy: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def isamax(n: int, x: np.ndarray, incx: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def idamax(n: int, x: np.ndarray, incx: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def isamin(n: int, x: np.ndarray, incx: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def idamin(n: int, x: np.ndarray, incx: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsadd(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdadd(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vssub(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdsub(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vssqr(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdsqr(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsmul(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdmul(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsabs(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdabs(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsfmod(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdfmod(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsremainder(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdremainder(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsinv(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdinv(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsdiv(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vddiv(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vssqrt(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdsqrt(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vspow(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdpow(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vspowx(n: int, a: np.ndarray, b: np.single, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdpowx(n: int, a: np.ndarray, b: np.single, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsexp(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdexp(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsexp2(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdexp2(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsexp10(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdexp10(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsln(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdln(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vslog2(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdlog2(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vslog10(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdlog10(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vserf(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vderf(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vserfc(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vderfc(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vscdfnorm(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdcdfnorm(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vserfinv(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vderfinv(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vserfcinv(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vderfcinv(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vscdfnorminv(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdcdfnorminv(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vslgamma(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdlgamma(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vstgamma(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdtgamma(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsexpint1(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdexpint1(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsfloor(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdfloor(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsceil(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdceil(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vstrunc(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdtrunc(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsround(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdround(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsnearbyint(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdnearbyint(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsrint(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdrint(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsmodf(n: int, a: np.ndarray, y: np.ndarray, z: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdmodf(n: int, a: np.ndarray, y: np.ndarray, z: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsfrac(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdfrac(n: int, a: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsfmax(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdfmax(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsfmin(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdfmin(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsmaxmag(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdmaxmag(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vsminmag(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def vdminmag(n: int, a: np.ndarray, b: np.ndarray, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def sparse_s_mv(operation: int, m: int, k: int, alpha: np.single, val: np.ndarray, indx: np.ndarray,
                    pntrb: np.ndarray, pntre: np.ndarray, descr, x: np.ndarray, beta: np.single,
                    y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def sparse_d_mv(operation: int, m: int, k: int, alpha: np.double, val: np.ndarray, indx: np.ndarray,
                    pntrb: np.ndarray, pntre: np.ndarray, descr, x: np.ndarray, beta: np.double,
                    y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def sparse_s_mm(operation: int, m: int, n: int, k: int, alpha: np.single, descr, layout: int, val: np.ndarray,
                    indx: np.ndarray, pntrb: np.ndarray, pntre: np.ndarray, x: np.ndarray, ldx: int, beta: np.single,
                    y: np.ndarray, ldy: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def sparse_d_mm(operation: int, m: int, n: int, k: int, alpha: np.double, descr, layout: int, val: np.ndarray,
                    indx: np.ndarray, pntrb: np.ndarray, pntre: np.ndarray, x: np.ndarray, ldx: int, beta: np.double,
                    y: np.ndarray, ldy: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def sparse_s_spmm(operation, request, sort, m, n, k, a, ja, ia, b, jb, ib, c, jc, ic, nzmax, info):
        pass

    @staticmethod
    @abc.abstractmethod
    def sparse_d_spmm(operation, request, sort, m, n, k, a, ja, ia, b, jb, ib, c, jc, ic, nzmax, info):
        pass

    @staticmethod
    @abc.abstractmethod
    def sparse_s_add(operation, request, sort, m, n, a, ja, ia, beta, b, jb, ib, c, jc, ic, nzmax, info):
        pass

    @staticmethod
    @abc.abstractmethod
    def sparse_d_add(operation, request, sort, m, n, a, ja, ia, beta, b, jb, ib, c, jc, ic, nzmax, info):
        pass

    @staticmethod
    @abc.abstractmethod
    def scsrmv(transa: int, m: int, k: int, alpha: np.single, matdescra: np.ndarray, val: np.ndarray, indx: np.ndarray,
               pntrb: np.ndarray, pntre: np.ndarray, x: np.ndarray, beta: np.single, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def dcsrmv(transa: int, m: int, k: int, alpha: np.double, matdescra: np.ndarray, val: np.ndarray, indx: np.ndarray,
               pntrb: np.ndarray, pntre: np.ndarray, x: np.ndarray, beta: np.double, y: np.ndarray):
        pass

    @staticmethod
    @abc.abstractmethod
    def scsrmm(transa: np.char, m: int, n: int, k: int, alpha: np.single, matdescra: np.ndarray, val: np.ndarray,
               indx: np.ndarray, pntrb: np.ndarray, pntre: np.ndarray, b: np.ndarray, ldb: int, beta: np.single,
               c: np.ndarray, ldc: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def dcsrmm(transa: np.char, m: int, n: int, k: int, alpha: np.double, matdescra: np.ndarray, val: np.ndarray,
               indx: np.ndarray, pntrb: np.ndarray, pntre: np.ndarray, b: np.ndarray, ldb: int, beta: np.double,
               c: np.ndarray, ldc: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def scsrmult(trans: np.char, request: int, sort: int, m: int, n: int, k: int, a: np.ndarray, ja: np.ndarray,
                 ia: np.ndarray, b: np.ndarray, jb: np.ndarray, ib: np.ndarray, c: np.ndarray, jc: np.ndarray,
                 ic: np.ndarray, nzmax: int, info: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def dcsrmult(trans: np.char, request: int, sort: int, m: int, n: int, k: int, a: np.ndarray, ja: np.ndarray,
                 ia: np.ndarray, b: np.ndarray, jb: np.ndarray, ib: np.ndarray, c: np.ndarray, jc: np.ndarray,
                 ic: np.ndarray, nzmax: int, info: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def scsradd(trans: np.char, request: int, sort: int, m: int, n: int, a: np.ndarray, ja: np.ndarray, ia: np.ndarray,
                beta: np.single, b: np.ndarray, jb: np.ndarray, ib: np.ndarray, c: np.ndarray, jc: np.ndarray,
                ic: np.ndarray, nzmax: int, info: int):
        pass

    @staticmethod
    @abc.abstractmethod
    def dcsradd(trans: np.char, request: int, sort: int, m: int, n: int, a: np.ndarray, ja: np.ndarray, ia: np.ndarray,
                beta: np.double, b: np.ndarray, jb: np.ndarray, ib: np.ndarray, c: np.ndarray, jc: np.ndarray,
                ic: np.ndarray, nzmax: int, info: int):
        pass



"""
This module includes AfeModellingSettings class.
"""
from typing import TypeVar, Generic

from organon.afe.domain.common.reader_helper import get_values_from_kwargs, get_val_from_kwargs
from organon.afe.domain.enums.afe_learning_type import AfeLearningType
from organon.afe.domain.modelling.supervised.afe_supervised_algorithm_settings import AfeSupervisedAlgorithmSettings
from organon.afe.domain.modelling.unsupervised.afe_unsupervised_algorithm_settings import \
    AfeUnsupervisedAlgorithmSettings
from organon.afe.domain.settings.afe_algorithm_settings import AfeAlgorithmSettings
from organon.afe.domain.settings.afe_output_settings import AfeOutputSettings
from organon.afe.domain.settings.afe_process_settings import AfeProcessSettings
from organon.afe.domain.settings.base_afe_data_settings import BaseAfeDataSettings
from organon.afe.domain.settings.base_afe_settings import BaseAfeSettings
from organon.afe.domain.settings.record_source import RecordSource

DS = TypeVar("DS", bound=BaseAfeDataSettings)
OS = TypeVar("OS", bound=AfeOutputSettings)


class BaseAfeModellingSettings(Generic[DS, OS], BaseAfeSettings):
    """
    Modelling Settings for Automated Feature Extraction process
    """
    ATTR_DICT = {
        "process_settings": AfeProcessSettings,
        "afe_learning_type": AfeLearningType,
        "data_source_settings": BaseAfeDataSettings,
        "output_settings": AfeOutputSettings,
        "algorithm_settings": AfeAlgorithmSettings,
        "scoring_target_source": RecordSource,
        "scoring_raw_input_source": RecordSource,
    }

    def __init__(self, **kwargs):
        super().__init__()
        self.output_settings: OS = None
        self.scoring_target_source: RecordSource = None
        self.scoring_raw_input_source: RecordSource = None
        self.afe_learning_type: AfeLearningType = None
        self.algorithm_settings: AfeAlgorithmSettings = None
        self.data_source_settings: DS = None

        get_values_from_kwargs(self, self.ATTR_DICT, kwargs,
                               ["algorithm_settings"])

        if "algorithm_settings" in kwargs:
            if self.afe_learning_type in [AfeLearningType.Supervised.name, AfeLearningType.Supervised]:
                self.algorithm_settings = get_val_from_kwargs(kwargs["algorithm_settings"],
                                                              AfeSupervisedAlgorithmSettings)
            elif self.afe_learning_type in [AfeLearningType.Unsupervised.name, AfeLearningType.Unsupervised]:
                self.algorithm_settings = get_val_from_kwargs(kwargs["algorithm_settings"],
                                                              AfeUnsupervisedAlgorithmSettings)
            else:
                raise ValueError("Invalid afe_learning_type value")



"""Includes ImputationService class."""
from typing import List, Union, Dict, Type

import numpy as np
import pandas as pd
# this import must stay: https://github.com/scikit-learn/scikit-learn/issues/16833
# noinspection PyUnresolvedReferences
from sklearn.experimental import enable_iterative_imputer  # pylint: disable=unused-import
from sklearn.impute import IterativeImputer
from sklearn.impute import SimpleImputer

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.helpers.data_frame_helper import get_column_native_type
from organon.fl.logging.helpers.log_helper import LogHelper
from organon.ml.common.helpers.parameter_helper import get_params
from organon.ml.preprocessing.domain.objects.imputation_fit_output import ImputationFitOutput
from organon.ml.preprocessing.settings.enums.imputer_type import ImputerType
from organon.ml.preprocessing.settings.objects.imputation_settings import ImputationSettings


class ImputationService:
    """Service for null value imputation"""

    def __init__(self, settings: ImputationSettings):
        self.settings = settings
        self.fit_output: ImputationFitOutput = None
        self.numerical_col_list: List[str] = None
        self.categorical_col_list: List[str] = None

    def fit(self, data: pd.DataFrame) -> ImputationFitOutput:
        """Fits train data by categorical and numeric imputers based on whether a col is numeric or not"""
        if data is None or data.empty:
            raise ValueError("Imputer cannot be fit. Data is not provided.")
        if self.settings.categorical_data_method == ImputerType.ITERATIVE:
            raise ValueError("IterativeImputer cannot be used for categorical columns")
        self._fill_column_lists(data)
        fit_output = ImputationFitOutput()
        fit_output.numerical_imputer = {}
        fit_output.categorical_imputer = {}
        categorical_imputer = self._get_imputer(self.settings.categorical_data_method)
        categorical_imputer_params = self._get_params(self.settings.categorical_data_method,
                                                      self.settings.c_strategy, self.settings.c_missing_values,
                                                      self.settings.c_fill_value)
        for col in self.categorical_col_list:
            imputer = categorical_imputer(**categorical_imputer_params)
            imputer.fit(data[[col]])
            fit_output.categorical_imputer[col] = imputer
        numerical_imputer = self._get_imputer(self.settings.numeric_data_method)
        numerical_imputer_params = self._get_params(self.settings.numeric_data_method, self.settings.n_strategy,
                                                    self.settings.n_missing_values, self.settings.n_fill_value)

        if self.settings.numeric_data_method == ImputerType.SIMPLE:
            for col in self.numerical_col_list:
                imputer = numerical_imputer(**numerical_imputer_params)
                imputer.fit(data[[col]])
                fit_output.numerical_imputer[col] = imputer
        else:
            imputer = numerical_imputer(**numerical_imputer_params)
            imputer.fit(data[self.numerical_col_list])
            fit_output.numerical_imputer = imputer
        self.fit_output = fit_output
        return fit_output

    @staticmethod
    def _get_imputer(imputing_method: ImputerType) -> Union[Type[SimpleImputer], Type[IterativeImputer]]:
        if imputing_method is ImputerType.SIMPLE:
            return SimpleImputer
        return IterativeImputer

    @staticmethod
    def _get_params(imputing_method: ImputerType, strategy: str,
                    missing_values: Union[int, float, str, None],
                    fill_value: Union[float, str, int]) -> Dict:
        if imputing_method is ImputerType.SIMPLE:
            return get_params({"strategy": strategy, "missing_values": missing_values, "fill_value": fill_value})
        return get_params({"initial_strategy": strategy, "missing_values": missing_values})

    def transform(self, data: pd.DataFrame, inplace: bool = False,
                  ignore_extra_columns: bool = False) -> pd.DataFrame:
        """Fills missing values in data and return it."""
        if self.fit_output is None:
            raise ValueError("Imputer not fitted yet. Run fit method first.")

        self._validate_transformation_data(data, ignore_extra_columns)

        if not inplace:
            trans_data = data.copy()
        else:
            trans_data = data
        categorical_col_list = [col for col in trans_data.columns if col in self.categorical_col_list]
        numerical_col_list = [col for col in trans_data.columns if col in self.numerical_col_list]

        for col in categorical_col_list:
            self._try_set_col_after_transform(trans_data, col, self.fit_output.categorical_imputer[col])
        if isinstance(self.fit_output.numerical_imputer, IterativeImputer):
            trans_data[self.numerical_col_list] = self.fit_output.numerical_imputer.transform(
                trans_data[self.numerical_col_list])
        else:
            for col in numerical_col_list:
                self._try_set_col_after_transform(trans_data, col, self.fit_output.numerical_imputer[col])
        return trans_data

    def _validate_transformation_data(self, trans_data: pd.DataFrame, ignore_extra_columns: bool):
        self._check_extra_columns_in_transformation_data(trans_data, ignore_extra_columns)

        if isinstance(self.fit_output.numerical_imputer, IterativeImputer):
            missing_cols = [col for col in self.numerical_col_list if col not in trans_data.columns]
            if missing_cols:
                raise ValueError(f"Following numerical columns should be supplied to transform data: "
                                 f"{self._get_list_str(missing_cols)}")

        self._check_changed_type_columns(trans_data)

    def _check_extra_columns_in_transformation_data(self, trans_data: pd.DataFrame, ignore_extra_columns: bool):
        train_cols = self.numerical_col_list + self.categorical_col_list
        if not ignore_extra_columns:
            different_columns = [col for col in trans_data.columns if col not in train_cols]
            if different_columns:
                raise ValueError(f"Imputer was not fitted for following columns: "
                                 f"{self._get_list_str(different_columns)}. "
                                 "Set 'ignore_extra_columns=True' if you want to transform only columns fitted "
                                 "and ignore others")

    def _check_changed_type_columns(self, trans_data: pd.DataFrame):
        numerical_changed_type_cols = []
        for col in self.numerical_col_list:
            if col in trans_data.columns:
                if get_column_native_type(trans_data, col) != ColumnNativeType.Numeric:
                    numerical_changed_type_cols.append(col)
        if numerical_changed_type_cols:
            raise ValueError(
                f"Following columns are expected to be numerical: {self._get_list_str(numerical_changed_type_cols)}")
        categorical_changed_type_cols = []
        for col in self.categorical_col_list:
            if col in trans_data.columns:
                if get_column_native_type(trans_data, col) != ColumnNativeType.String:
                    categorical_changed_type_cols.append(col)
        if categorical_changed_type_cols:
            raise ValueError(
                f"Following columns are expected to be categorical: "
                f"{self._get_list_str(categorical_changed_type_cols)}")

    @classmethod
    def _try_set_col_after_transform(cls, trans_data: pd.DataFrame, col: str, imputer):
        initial_dtype = trans_data[col].dtype
        transformed_arr = imputer.transform(trans_data[[col]])
        if transformed_arr.shape[1] != 0:
            trans_data[col] = transformed_arr
            if trans_data[col].dtype == np.dtype("O"):
                # 'imputer.transform' can convert types like Int64DType to object. Here we try to convert it to
                try:
                    trans_data[col] = trans_data[col].astype(initial_dtype)
                except:  # pylint: disable=bare-except
                    trans_data[col] = trans_data[col].convert_dtypes()
        else:
            if trans_data[col].isna().all():
                LogHelper.warning(f"Column {col} could not be transformed by imputer since all values are NaN.")

    def _fill_column_lists(self, data: pd.DataFrame):
        self.numerical_col_list = []
        self.categorical_col_list = []
        if self.settings.included_columns is None:
            columns = data.columns
        else:
            columns = self.settings.included_columns
        for col in columns:
            col_type = get_column_native_type(data, col)
            if col_type == ColumnNativeType.String:
                self.categorical_col_list.append(col)
            elif col_type == ColumnNativeType.Numeric:
                self.numerical_col_list.append(col)

    @staticmethod
    def _get_list_str(_list: list):
        return ",".join(_list)



"""Includes ISampleDataService interface"""
import abc

import pandas as pd


class ISampleDataService(metaclass=abc.ABCMeta):
    """Interface for SampleData services."""

    @abc.abstractmethod
    def get_sample(self) -> pd.DataFrame:
        """Read and return sample as pd.DataFrame"""



"""Includes BaseSupervisedFeatureSelectionSettings class."""
import pandas as pd

from organon.ml.common.enums.target_type import TargetType
from organon.ml.feature_selection.domain.objects.settings.base_feature_selection_settings import \
    BaseFeatureSelectionSettings


class BaseSupervisedFeatureSelectionSettings(BaseFeatureSelectionSettings):
    """Base settings for supervised feature selection services"""

    def __init__(self, data: pd.DataFrame, target: pd.DataFrame, target_type: TargetType):
        super().__init__(data)
        self.target: pd.DataFrame = target
        self.target_type: TargetType = target_type



"""Module for LinearRegressionResults"""
import math
from dataclasses import dataclass


@dataclass
class LinearRegressionResults:
    """Statistical results of the linear regression algorithm"""
    parameters = None
    sample_size = None
    bias_dict = {}
    r_square = None
    adjusted_r_square = None
    model_degrees_of_freedom = None
    error_degrees_of_freedom = None
    total_sum_of_squares = None
    explained_sum_of_squares = None
    residual_sum_of_squares = None
    mean_square_error = None
    mean_square_regression = None
    f_statistics = None
    f_statistics_p_value = None

    # pylint: disable=unsubscriptable-object
    def __str__(self) -> str:
        result_string = f"Number of samples: {self.sample_size}\n"
        result_string += f"RSquared: {self.r_square}\n"
        result_string += f"Adjusted-RSquared: {self.adjusted_r_square}\n"
        result_string += f"Standard Error-of-Regression: {math.sqrt(self.mean_square_error)}\n"
        result_string += "\n"
        result_string += "\n"
        result_string += f"Number of samples: {self.sample_size}\n"
        result_string += "\n"
        result_string += "Analysis of Variance:\n"
        result_string += "\n"
        result_string += f"               {'DF':<15}{'Sum of Squares':<25}{'Mean Square':<25}\n"
        result_string += f"{'Regression':<15}{self.model_degrees_of_freedom:<15,G}" \
                         f"{self.explained_sum_of_squares:<25,E}{self.mean_square_regression:<25,E}\n"
        result_string += f"{'Residuals':<15}{(self.sample_size- self.model_degrees_of_freedom - 1):<15,G}" \
                         f"{self.residual_sum_of_squares:<25,E}{self.mean_square_error:<25,E}\n"
        result_string += "\n"
        result_string += "\n"
        result_string += "\n"
        result_string += f"{'NAME':<25}{'B':<20}{'SEB':<20}{'T':<20}{'VIF':<20}\n"
        result_string += f"{'CONSTANT':<25}{round(self.bias_dict['estimate'],6):<20}{float('nan'):<20}" \
                         f"{float('nan'):<20}{float('nan'):<20}\n"

        for key in self.parameters:  # pylint: disable=not-an-iterable
            result_string += f"{key:<25}{round(self.parameters[key].estimate, 6):<20}" \
                             f"{ round(self.parameters[key].standard_error,6):<20}" \
                             f"{round(self.parameters[key].test_statistic, 6):<20}" \
                             f"{round(self.parameters[key].vif, 6):<20}\n"

        return result_string



"""
This module includes AfeDataSettings class.
"""
from typing import List, Generic, TypeVar

from organon.afe.domain.common.reader_helper import get_values_from_kwargs
from organon.afe.domain.enums.record_source_type import RecordSourceType
from organon.afe.domain.settings.auto_column_decider_settings import AutoColumnDeciderSettings
from organon.afe.domain.settings.base_trx_descriptor import BaseTrxDescriptor
from organon.afe.domain.settings.record_source import RecordSource
from organon.afe.domain.settings.target_descriptor import TargetDescriptor

T = TypeVar("T", bound=BaseTrxDescriptor)


class BaseAfeDataSettings(Generic[T]):
    """
    Stores data source settings for AfeModellingSettings
    """
    ATTR_DICT = {
        "trx_descriptor": BaseTrxDescriptor,
        "target_descriptor": TargetDescriptor,
        "auto_column_decider_settings": AutoColumnDeciderSettings,
        "target_record_source_list": List[RecordSource],
        "max_number_of_target_samples": int,
        "max_number_of_transaction_samples": int,
    }

    def __init__(self, **kwargs):
        self.trx_descriptor: T = None
        self.target_descriptor: TargetDescriptor = None
        self.auto_column_decider_settings: AutoColumnDeciderSettings = None
        self.target_record_source_list: List[RecordSource] = None
        self.max_number_of_target_samples: int = None
        self.max_number_of_transaction_samples: int = None

        get_values_from_kwargs(self, self.ATTR_DICT, kwargs)

    def is_homogenous(self) -> bool:
        """checks if all target record sources have the same connection name as the transaction record source"""
        trx_source = self.trx_descriptor.modelling_raw_input_source
        if trx_source.get_type() != RecordSourceType.DATABASE:
            return False
        for record_source in self.target_record_source_list:  # pylint: disable=not-an-iterable
            if record_source.get_type() != RecordSourceType.DATABASE or \
                    record_source.source.connection_name != trx_source.source.connection_name:
                return False
        return True



"""Includes DqSampleNumericalStatistics class."""
from typing import List

from organon.fl.core.collections.sorted_dict import SortedDict
from organon.fl.mathematics.businessobjects.sets.interval import Interval
from organon.fl.mathematics.businessobjects.sets.sum_triple import SumTriple
from organon.idq.domain.businessobjects.statistics.dq_base_statistics import DqBaseStatistics
from organon.idq.domain.helpers.statistics.dq_statistical_functions import DqStatisticalFunctions


class DqSampleNumericalStatistics(DqBaseStatistics):
    """todo"""

    # pylint: disable=too-many-instance-attributes

    def __init__(self):
        super().__init__()
        self.missing_values: List[float] = None
        self.missing_values_frequencies: SortedDict[float, int] = None
        self.compute_frequencies: bool = None
        self.frequencies: SortedDict[float, float] = None
        self.interval_statistics: SortedDict[Interval, SumTriple] = None
        self.mean: float = None
        self.trimmed_mean: float = None
        self.std_dev: float = None
        self.variance: float = None
        self.min: float = None
        self.max: float = None
        self.sum: float = None
        self.median: float = None
        self.inter_quartile_range: float = None
        self.tukey_lower_limit: float = None
        self.tukey_upper_limit: float = None
        self.percentile_values: SortedDict[int, float] = None
        self.quartile_values: SortedDict[int, float] = None

    @property
    def percentages(self):
        """todo"""
        return DqStatisticalFunctions.get_percentages_from_frequencies_for_sorted(self.frequencies)

    @staticmethod
    def get_null_statistics():
        """Return DqSampleNumericalStatistics with null statistics."""
        stats = DqSampleNumericalStatistics()
        stats.n_val = 0
        stats.n_miss = 0
        stats.cardinality = 0
        stats.compute_frequencies = False
        stats.mean = float("nan")
        stats.trimmed_mean = float("nan")
        stats.std_dev = float("nan")
        stats.variance = float("nan")
        stats.min = float("nan")
        stats.max = float("nan")
        stats.sum = float("nan")
        stats.median = float("nan")
        stats.inter_quartile_range = float("nan")
        stats.tukey_lower_limit = float("nan")
        stats.tukey_upper_limit = float("nan")
        return stats



"""Includes ObjectDetectionTrainer class"""
import os
import urllib

import numpy as np
import yaml

from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.core.fileoperations import directory_helper
from organon.fl.core.helpers import guid_helper
from organon.fl.logging.helpers.log_helper import LogHelper
from organon.ml.common.helpers.user_input_service_helper import get_enum
from organon.ml.object_detection.domain.object_detection_training_service import ObjectDetectionTrainingService
from organon.ml.object_detection.services.enums.map_metric import MapMetrics
from organon.ml.object_detection.services.enums.model_type import ModelTypes


class ObjectDetectionTrainer:
    """Class for training ObjectDetection models"""

    def __init__(self, pretrained_model_path: str, yaml_data_dir: str, batch_size: int = 16,
                 epochs: int = 100, map_metric: str = 'map50', map_threshold: float = 0.8):
        map_metric = get_enum(map_metric.upper(), MapMetrics)

        self._service: ObjectDetectionTrainingService = None
        self._yaml_data_dir = yaml_data_dir
        self._pretrained_model_path = pretrained_model_path
        self._batch_size = batch_size
        self._epochs = epochs
        self._is_trained_on_full_data = False
        self._map_metric: MapMetrics = map_metric
        self._map_threshold: float = map_threshold

    def train(self, sampling: bool = False, sampling_ratio: float = 0.3, random_seed: int = 42,
              output_folder: str = None, override_data_files: bool = False):
        """
        Runs training
        :param sampling: If True, training will be done on a sample of the train data.
            If best mAP value among models created during this training is higher than
            pre_train_eval_min_map(see set_pre_train_settings method), training on full data will not run. Else,
            training will run again on full data
        :param sampling_ratio:
        :param random_seed:
        :param output_folder:
        :param override_data_files:
        """

        if output_folder is None:
            output_folder = os.path.dirname(self._yaml_data_dir)

        if not directory_helper.exists(output_folder):
            raise ValueError(f"Directory '{output_folder}' does not exist")
        train_on_full_data_flag = True
        # pylint: disable=too-many-arguments
        if sampling:
            LogHelper.warning("Training with sample data started")
            new_yaml_dir = None
            try:
                new_yaml_dir = self._generate_new_data_dir_for_pre_train(sampling_ratio, random_seed)

                pre_train_service = ObjectDetectionTrainingService(new_yaml_dir,
                                                                   self._pretrained_model_path)
                pre_train_service.train(batch_size=self._batch_size, epochs=self._epochs, random_seed=random_seed,
                                        output_folder=output_folder, override_data_files=override_data_files)

                map_score = pre_train_service.evaluate(self._map_metric)
                if map_score > self._map_threshold:
                    train_on_full_data_flag = False
                    LogHelper.warning(
                        f"map score is {map_score} which is greater than the threshold {self._map_threshold}."
                        "Training on full data step will be skipped.")
                    self._service = pre_train_service
                    self._is_trained_on_full_data = train_on_full_data_flag
                    return
                directory_helper.delete_directory_with_contents(os.path.dirname(new_yaml_dir))

            except Exception as exc:
                LogHelper.error("Error occurred while pre-training")
                if new_yaml_dir:
                    LogHelper.info("Cleaning data directory generated for pre-training")
                    directory_helper.delete_directory_with_contents(os.path.dirname(new_yaml_dir))

                raise exc

        try:
            if train_on_full_data_flag:
                service = ObjectDetectionTrainingService(self._yaml_data_dir, self._pretrained_model_path)
                LogHelper.info("Training with full data started")
                service.train(batch_size=self._batch_size, epochs=self._epochs, random_seed=random_seed,
                              output_folder=output_folder, override_data_files= override_data_files)
                self._service = service
                self._is_trained_on_full_data = train_on_full_data_flag
        except Exception as exc:
            msg = "Error occurred while training on full data."
            raise Exception(msg) from exc

    @property
    def is_trained_on_full_data(self):
        """Returns if the model is trained on full data, not pretrain data"""
        return self._is_trained_on_full_data

    @classmethod
    def load_pretrained_model_from_web(cls, target_path: str,
                                       model_type: str = "small"):
        """Loads a pretrained model to given file path after downloading from web. The default model to be downloaded is
        'YOLOv8s' and its approx 25 MB
        :param target_path: File path to download the pretrained model in
        :param model_type: Model type keyword to be downloaded
        """
        model_type_enum = get_enum(model_type.upper(), ModelTypes)
        model_url = ObjectDetectionTrainer._get_model_by_model_type(model_type_enum)

        if os.path.isfile(target_path):
            raise FileExistsError(f"File '{target_path}' already exists")
        try:
            LogHelper.warning(f"Started loading pretrained model from '{model_url}'")
            urllib.request.urlretrieve(model_url, target_path)  # nosec
        except Exception as exc:
            msg = "Error occurred while loading a pretrained model from internet"
            raise Exception(msg) from exc
        LogHelper.warning(f"Pretrained model loaded to file {target_path}")

    def _generate_new_data_dir_for_pre_train(self, data_sampling_ratio: float, random_seed: int):
        guid = guid_helper.new_guid(8)
        try:
            new_data_dir = os.path.join(os.path.dirname(self._yaml_data_dir),
                                        f"sample_{guid}")
            LogHelper.info(f"Generating new data directory for pre training in {new_data_dir}")
            directory_helper.create(new_data_dir, exception_if_exists=True)
            original_train_image_dir = ObjectDetectionTrainingService.get_train_images_dir_path_yaml(
                self._yaml_data_dir)
            original_train_labels_dir = ObjectDetectionTrainingService.get_train_label_dir_path_yaml(
                self._yaml_data_dir)
            valid_image_files_dir = ObjectDetectionTrainingService.get_validation_images_dir_path_yaml(
                self._yaml_data_dir)
            valid_labels_dir = ObjectDetectionTrainingService.get_validation_label_dir_path_yaml(self._yaml_data_dir)

            train_image_files, train_label_files = self._get_sampled_image_files(original_train_image_dir,
                                                                                 original_train_labels_dir,
                                                                                 data_sampling_ratio, random_seed)
            valid_image_files, val_label_files = self._get_sampled_image_files(valid_image_files_dir, valid_labels_dir,
                                                                               data_sampling_ratio, random_seed)

            new_train_images_dir = ObjectDetectionTrainingService.get_train_images_dir_path(new_data_dir)
            new_train_labels_dir = ObjectDetectionTrainingService.get_train_label_dir_path(new_data_dir)
            new_valid_images_dir = ObjectDetectionTrainingService.get_validation_images_dir_path(new_data_dir)
            new_valid_labels_dir = ObjectDetectionTrainingService.get_validation_label_dir_path(new_data_dir)

            new_yaml_dir = ObjectDetectionTrainer. \
                _create_updated_yaml_file_with_new_data_locations(self._yaml_data_dir,
                                                                  new_data_dir,
                                                                  new_train_images_dir,
                                                                  new_valid_images_dir)

            LogHelper.info(f"Copying sampled train images for pre training in {new_train_images_dir}")
            directory_helper.copy_files(train_image_files, new_train_images_dir)
            LogHelper.info(f"Copying sampled validation images for pre training in {new_valid_images_dir}")
            directory_helper.copy_files(valid_image_files, new_valid_images_dir)
            LogHelper.info(f"Copying sampled validation labels for pre training in {new_valid_labels_dir}")
            directory_helper.copy_files(val_label_files, new_valid_labels_dir)
            LogHelper.info(f"Copying sampled train images for pre training in {new_train_labels_dir}")
            directory_helper.copy_files(train_label_files, new_train_labels_dir)
        except KnownException as k_exc:
            raise k_exc
        except ValueError as v_exc:
            raise v_exc
        except Exception as exc:
            msg = "Error occurred while generating new data dirs for sampling."
            raise Exception(msg) from exc

        return new_yaml_dir

    @staticmethod
    def _get_sampled_image_files(original_image_files_dir: str, original_label_files_dir: str,
                                 data_sampling_ratio: float, random_seed: int):
        all_file_names = os.listdir(original_image_files_dir)
        num_files_to_get = int(len(all_file_names) * data_sampling_ratio)
        if num_files_to_get == 0:
            raise KnownException("Not enough files to run pre-evaluation, "
                                 "try increasing pre_train_eval_data_sampling_ratio or disable pre evaluation "
                                 "to run training directly with full data")
        np.random.seed(random_seed)
        files = np.random.choice(all_file_names, num_files_to_get, replace=False)
        image_files = [os.path.join(original_image_files_dir, file) for file in files]
        label_files = [os.path.join(original_label_files_dir,
                                    f"{os.path.splitext(os.path.basename(file))[0]}.txt") for file in image_files]

        return image_files, label_files

    @staticmethod
    def _create_updated_yaml_file_with_new_data_locations(yaml_file_location, new_data_dir, new_train_dir, new_val_dir):

        yaml_dir, yaml_filename = os.path.split(yaml_file_location)

        updated_yaml_filename = yaml_filename.replace('.yaml', '_updated.yaml')

        relative_path = os.path.relpath(new_data_dir, yaml_dir)
        updated_yaml_dir = os.path.join(yaml_dir, relative_path)

        updated_yaml_file_location = os.path.join(updated_yaml_dir, updated_yaml_filename)

        os.makedirs(updated_yaml_dir, exist_ok=True)

        with open(yaml_file_location, encoding="utf-8") as file:
            yaml_data = yaml.safe_load(file)

        updated_yaml_data = yaml_data.copy()
        updated_yaml_data['train'] = new_train_dir
        updated_yaml_data['val'] = new_val_dir

        with open(updated_yaml_file_location, 'w', encoding="utf-8") as file:
            yaml.dump(updated_yaml_data, file)

        return updated_yaml_file_location

    @staticmethod
    def _get_model_by_model_type(model_type: ModelTypes):
        """return the model url from given model keyword"""

        if model_type == ModelTypes.NANO:
            model_url = "https://github.com/ultralytics/assets/releases/download/v0.0.0/yolov8n.pt"
        elif model_type == ModelTypes.SMALL:
            model_url = "https://github.com/ultralytics/assets/releases/download/v0.0.0/yolov8s.pt"
        elif model_type == ModelTypes.MEDIUM:
            model_url = "https://github.com/ultralytics/assets/releases/download/v0.0.0/yolov8m.pt"
        elif model_type == ModelTypes.LARGE:
            model_url = "https://github.com/ultralytics/assets/releases/download/v0.0.0/yolov8l.pt"
        elif model_type == ModelTypes.XLARGE:
            model_url = "https://github.com/ultralytics/assets/releases/download/v0.0.0/yolov8x.pt"
        else:
            raise ValueError("Wrong Model Type keyword. Model Types can only be nano, small, medium, large and Xlarge")
        return model_url



"""Includes CoarseClassSettings class."""
from dataclasses import dataclass

from organon.ml.common.enums.target_type import TargetType


@dataclass
class CoarseClassSettings:
    """Settings for CoarseClass service"""
    test_ratio: float
    min_class_size: int
    target_type: TargetType
    max_leaf_nodes: int
    stability_check: bool
    stability_threshold: float
    random_state: int
    positive_class: str
    negative_class: str
